import copy
import xml.etree.ElementTree as ET
from xml.dom import minidom
import os
import binascii


def modifyScenarioGeneralSettings(scenXml, parameterDict):
    print("---Modifying " + scenXml)
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # PARSE SCENARIO GENERAL SETTINGS, AND MODIFY THEM ACCORDING TO THE PROVIDED
    # PARAMETER LIST
    for settings in root.findall('Settings'):
        for setting in settings.findall('Setting'):
            for param in parameterDict:
                if param == setting.find('Name').text:
                    xmlVal = setting.find('Value')
                    xmlVal.text = parameterDict[param]

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyExtractionIO(scenXml, newFilename, newOutputSpect1, newOutputSpect2, newOutputBaseline1, newOutputBaseline2, newOutputConnect1, newOutputConnect2, newOutputTrials, newExtractIdx):
    print("---Modifying " + scenXml + " input and output")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # PARSE SCENARIO GENERAL SETTINGS, AND MODIFY THEM ACCORDING TO THE PROVIDED
    # PARAMETER LIST
    for settings in root.findall('Settings'):
        for setting in settings.findall('Setting'):
            if setting.find('Name').text == "EEGData":
                xmlVal = setting.find('Value')
                xmlVal.text = newFilename
            elif setting.find('Name').text == "OutputSpect1":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputSpect1
            elif setting.find('Name').text == "OutputSpect2":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputSpect2
            elif setting.find('Name').text == "OutputConnect1":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputConnect1
            elif setting.find('Name').text == "OutputConnect2":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputConnect2
            elif setting.find('Name').text == "OutputBaseline1":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputBaseline1
            elif setting.find('Name').text == "OutputBaseline2":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputBaseline2
            elif setting.find('Name').text == "OutputTrials":
                xmlVal = setting.find('Value')
                xmlVal.text = newOutputTrials
            elif setting.find('Name').text == "ExtractionIdx":
                xmlVal = setting.find('Value')
                xmlVal.text = newExtractIdx

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyAcqScenario(scenXml, parameterDict, boolOnline):
    print("---Modifying " + scenXml + " Graz Variables")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # PARSE SCENARIO BOX SETTINGS, AND MODIFY GRAZ PROTOCOL VARIABLES
    # PARAMETER LIST
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == 'Graz Motor Imagery BCI Stimulator':
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Number of Trials for Each Class":
                            xmlVal = setting.find('Value')
                            xmlVal.text = parameterDict["TrialNb"]
                            continue
                        elif setting.find('Name').text == "Baseline Duration (in sec)":
                            xmlVal = setting.find('Value')
                            xmlVal.text = parameterDict["Baseline"]
                            continue
                        elif setting.find('Name').text == "Wait For Cue Duration (in sec)":
                            xmlVal = setting.find('Value')
                            xmlVal.text = parameterDict["TrialWait"]
                            continue
                        elif setting.find('Name').text == "Display Cue Duration (in sec)":
                            xmlVal = setting.find('Value')
                            xmlVal.text = parameterDict["TrialLength"]
                            continue
                        elif setting.find('Name').text == "End of Trial Minimum Duration (in sec)":
                            xmlVal = setting.find('Value')
                            xmlVal.text = parameterDict["EndTrialMin"]
                            continue
                        elif setting.find('Name').text == "End of Trial Maximum Duration (in sec)":
                            xmlVal = setting.find('Value')
                            xmlVal.text = parameterDict["EndTrialMax"]
                            continue
                        elif setting.find('Name').text == "Feedback Duration (in sec)":
                            if boolOnline:
                                xmlVal = setting.find('Value')
                                xmlVal.text = parameterDict["FeedbackLength"]
                                continue
                            else:
                                xmlVal = setting.find('Value')
                                xmlVal.text = str(0)
                                continue

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyTrainScenario(chanFreqPairs, epochCount, scenXml):
    print("---Modifying " + scenXml + " with Selected Features")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # FIRST STEP AND
    # SIMPLEST CASE : modify existing branches in the scenario.
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):

            if box.find('Name').text == 'Channel Selector':
                print("-- CHANNEL SELECTOR BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Channel List":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = chanFreqPairs[0][0]
                            print("            with " + xmlVal.text)
                            break

            if box.find('Name').text == 'Frequency Band Selector':
                print("-- FREQ SELECTION BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        # POWER SPECTRUM PIPELINE: actual "freq band selector" box
                        # CONNECTIVITY PIPELINE: channel selector rebranded as freq band selector
                        if setting.find('Name').text == "Frequencies to select" or setting.find('Name').text == "Channel List":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = chanFreqPairs[0][1]
                            print("            with " + xmlVal.text)
                            break
            if box.find('Name').text == 'Epoch average':
                print("-- EPOCH AVERAGE BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Epoch count":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = str(epochCount)
                            print("            with " + xmlVal.text)
                            break

    if len(chanFreqPairs) > 1:
        # TOUGHEST CASE : NEED TO COPY/PASTE MULTIPLE PROCESSING
        # BRANCHES WITH CHANNEL SELECTOR, TIME EPOCH, AR COEFFS,
        # FREQ BAND SELECTOR AND SPECTRUM AVERAGE

        # FIRST STEP : GET THE TWO "SPLIT" IDENTITY BOXES
        splitBoxes = []
        splitIdx = 0
        for boxes in root.findall('Boxes'):
            for box in boxes.findall('Box'):
                if box.find('Name').text == 'SPLIT':
                    print("-- SPLIT " + str(splitIdx))
                    splitBoxes.append(box)
                    splitIdx += 1
                    continue

        # Find Id of Classifier trainer box
        boxLast = findBox(root, "Classifier trainer")
        boxLastId = boxLast.find("Identifier").text
        for box in splitBoxes:
            boxId = box.find("Identifier").text
            locOffset = 120

            # Find all chained box btw those two
            if boxId is not None and boxLastId is not None:
                boxList = findChainedBoxes(root, boxId, boxLastId)

                for idxPair, [chan, freq] in enumerate(chanFreqPairs):
                    # Don't do it for the first pair, it was done earlier in the function
                    if idxPair == 0:
                        continue

                    pair = [chan, freq]
                    # Copy list of chained boxes (except the first (SPLIT) and
                    # the 2 last ones (feature aggreg, classifier trainer) and chain them.
                    listofBoxesToChain, nbOfOutputs = copyBoxList(root, boxList, locOffset, pair, 1, 2)
                    locOffset += locOffset

                    # Add an input to Feature Aggregator box in the current chain
                    addInputToBox(root, listofBoxesToChain[-1])
                    featAggInputIdx = countBoxInputs(root, listofBoxesToChain[-1]) - 1
                    linkBoxes(root, listofBoxesToChain, nbOfOutputs, featAggInputIdx)

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyTrainScenUsingSplitAndClassifiers(splitStr, classifStr, chanFreqPairs, epochCount, scenXml):
    print("---Modifying " + scenXml + " with Selected Features")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # FIRST STEP : GET THE "SPLIT" IDENTITY BOXES
    splitBoxes = []
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == splitStr:
                splitBoxes.append(box)
                print("-- SPLIT POWSPECT " + str(len(splitBoxes)+1))
                continue

    # Find Id of Classifier trainer box
    boxLast = findBox(root, classifStr)
    boxLastId = boxLast.find("Identifier").text

    # FIRST STEP : modify existing branches in the scenario, using the first pair of features
    for splitbox in splitBoxes:
        boxId = splitbox.find("Identifier").text
        # Find all boxes chained btw this box and the last one (classifier trainer or processor)
        if boxId is not None and boxLastId is not None:
            boxList = findChainedBoxes(root, boxId, boxLastId)

        # Change parameters of particular boxes in this list of chained boxes
        for boxid in boxList:
            box = findBoxId(root, boxid)
            if box.find('Name').text == 'Channel Selector':
                print("-- CHANNEL SELECTOR BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Channel List":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = chanFreqPairs[0][0]
                            print("            with " + xmlVal.text)
                            break

            if box.find('Name').text == 'Frequency Band Selector':
                print("-- FREQ SELECTION BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        # POWER SPECTRUM PIPELINE: actual "freq band selector" box
                        # CONNECTIVITY PIPELINE: channel selector rebranded as freq band selector
                        if setting.find('Name').text == "Frequencies to select" or setting.find('Name').text == "Channel List":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = chanFreqPairs[0][1]
                            print("            with " + xmlVal.text)
                            break
            if box.find('Name').text == 'Epoch average':
                print("-- EPOCH AVERAGE BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Epoch count":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = str(epochCount)
                            print("            with " + xmlVal.text)
                            break

    if len(chanFreqPairs) > 1:
        # COPY/PASTE PROCESSING BRANCHES
        for splitbox in splitBoxes:
            locOffset = 120
            boxId = splitbox.find("Identifier").text
            # Find all boxes chained btw this split and the last (classifier trainer or processor)
            if boxId is not None and boxLastId is not None:
                boxList = findChainedBoxes(root, boxId, boxLastId)

                for idxPair, [chan, freq] in enumerate(chanFreqPairs):
                    # Don't do it for the first pair, it was done earlier in the function
                    if idxPair == 0:
                        continue

                    pair = [chan, freq]
                    # Copy list of chained boxes (except the first (SPLIT) and
                    # the 2 last ones (feature aggreg, classifier trainer or processor) and chain them.
                    listofBoxesToChain, nbOfOutputs = copyBoxList(root, boxList, locOffset, pair, 1, 2)

                    # Add an input to Feature Aggregator box in the current chain
                    addInputToBox(root, listofBoxesToChain[-1])
                    featAggInputIdx = countBoxInputs(root, listofBoxesToChain[-1]) - 1
                    linkBoxes(root, listofBoxesToChain, nbOfOutputs, featAggInputIdx)

                    locOffset += locOffset

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyTrainScenUsingSplitAndCsvWriter(splitStr, chanFreqPairs, epochCount, scenXml, trainingpath):
    print("---Modifying " + scenXml + " with Selected Features")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # FIRST STEP : GET THE "SPLIT" IDENTITY BOXES
    splitBoxes = []
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == splitStr:
                splitBoxes.append(box)
                print("-- SPLIT " + str(len(splitBoxes)+1))
                continue

    # Find Ids of CSV Writer boxes
    csvBoxes = []
    csvBoxesIds = []
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == "CSV Class1 Feat1":
                csvBoxes.append(box)
                csvBoxesIds.append(csvBoxes[-1].find("Identifier").text)
                print("-- CSV WRITER " + str(len(csvBoxes) + 1))
                continue
            elif box.find('Name').text == "CSV Class2 Feat1":
                csvBoxes.append(box)
                csvBoxesIds.append(csvBoxes[-1].find("Identifier").text)
                print("-- CSV WRITER " + str(len(csvBoxes) + 1))
                continue

    # FIRST STEP : modify existing branches in the scenario, using the first pair of features
    for splitbox in splitBoxes:
        boxId = splitbox.find("Identifier").text

        boxListFound = False
        # Find all boxes chained btw this box and the last one (csv Writer)
        for csvBoxId in csvBoxesIds:
            if boxId is not None and csvBoxId is not None:
                boxList = findChainedBoxes(root, boxId, csvBoxId)

                if boxList:
                    boxListFound = True
                    # Change parameters of particular boxes in this list of chained boxes
                    for boxid in boxList:
                        box = findBoxId(root, boxid)
                        if box.find('Name').text == 'Channel Selector':
                            print("-- CHANNEL SELECTOR BOX ")
                            for settings in box.findall('Settings'):
                                for setting in settings.findall('Setting'):
                                    if setting.find('Name').text == "Channel List":
                                        xmlVal = setting.find('Value')
                                        print("       replacing " + xmlVal.text)
                                        xmlVal.text = chanFreqPairs[0][0]
                                        print("            with " + xmlVal.text)
                                        break

                        if box.find('Name').text == 'Frequency Band Selector':
                            print("-- FREQ SELECTION BOX ")
                            for settings in box.findall('Settings'):
                                for setting in settings.findall('Setting'):
                                    # POWER SPECTRUM PIPELINE: actual "freq band selector" box
                                    # CONNECTIVITY PIPELINE: channel selector rebranded as freq band selector
                                    if setting.find('Name').text == "Frequencies to select" or setting.find('Name').text == "Channel List":
                                        xmlVal = setting.find('Value')
                                        print("       replacing " + xmlVal.text)
                                        xmlVal.text = chanFreqPairs[0][1]
                                        print("            with " + xmlVal.text)
                                        break
                        if box.find('Name').text == 'Epoch average':
                            print("-- EPOCH AVERAGE BOX ")
                            for settings in box.findall('Settings'):
                                for setting in settings.findall('Setting'):
                                    if setting.find('Name').text == "Epoch count":
                                        xmlVal = setting.find('Value')
                                        print("       replacing " + xmlVal.text)
                                        xmlVal.text = str(epochCount)
                                        print("            with " + xmlVal.text)
                                        break
        if not boxListFound:
            print("-- ERROR IN TEMPLATE SCENARIO !! --")
            return

    if len(chanFreqPairs) > 1:
        # COPY/PASTE PROCESSING BRANCHES
        for splitbox in splitBoxes:

            locOffset = 120
            boxId = splitbox.find("Identifier").text
            boxListFound = False

            # Find all boxes chained btw this box and the last one (csv Writer)
            for csvBoxId in csvBoxesIds:
                if boxId is not None and csvBoxId is not None:
                    boxList = findChainedBoxes(root, boxId, csvBoxId)

                    if boxList:
                        boxListFound = True

                        for idxPair, [chan, freq] in enumerate(chanFreqPairs):
                            # Don't do it for the first pair, it was done earlier in the function
                            if idxPair == 0:
                                continue

                            pair = [chan, freq]
                            # Copy list of chained boxes (except the first (SPLIT)) and chain them.
                            listofBoxesToChain, nbOfOutputs = copyBoxListGeneric(root, boxList, locOffset, pair, 1, 0)
                            csvWriter = findBoxId(root, listofBoxesToChain[-1])
                            oldName = csvWriter.find('Name').text.split("Feat")
                            changeBoxName(csvWriter, str(oldName[0] + 'Feat' + str(int(oldName[1])+ idxPair) ) )
                            linkBoxesGeneric(root, listofBoxesToChain, nbOfOutputs)

                            locOffset += locOffset

            if not boxListFound:
                print("-- ERROR IN TEMPLATE SCENARIO !! --")
                return

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyTrainPartitions(trainParts, scenXml):
    print("---Modifying " + scenXml + " with train partitions: " + str(trainParts))
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # FIRST STEP AND
    # SIMPLEST CASE : just modify existing branches in the scenario.
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == 'Classifier trainer':
                print("-- CLASSIFIER TRAINER BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Number of partitions for k-fold cross-validation test":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = str(trainParts)
                            print("            with " + xmlVal.text)
                            break

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyTrainIO(newFilename, newWeightsName, newExtractIdx, scenXml):
    print("---Modifying " + scenXml + " input")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # PARSE SCENARIO GENERAL SETTINGS, AND MODIFY THEM ACCORDING TO THE PROVIDED
    # PARAMETER LIST
    for settings in root.findall('Settings'):
        for setting in settings.findall('Setting'):
            if setting.find('Name').text == "EEGData":
                xmlVal = setting.find('Value')
                xmlVal.text = newFilename
            elif setting.find('Name').text == "OutputWeights":
                xmlVal = setting.find('Value')
                xmlVal.text = newWeightsName
            elif setting.find('Name').text == "ExtractionIdx":
                xmlVal = setting.find('Value')
                xmlVal.text = newExtractIdx

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyOnlineScenario(chanFreqPairs, scenXml):
    print("---Modifying " + scenXml + " with Selected Features")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # Assuming Power Spectrum pipeline...
    # There are two distinct scenario branches to modify here :
    # - Direct feedback branch : take the *first* electrode/freq pair
    # - Online Classification : behave exactly as for the training part (clone branches)

    channelSelectorBox = None
    # FIRST STEP AND
    # SIMPLEST CASE : just modify existing branches in the scenario.
    # (this takes care of the "direct feedback branch" + the first feature pair
    # of the "online classification" branch")
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):

            if box.find('Name').text == 'Channel Selector':
                print("-- CHANNEL SELECTOR BOX ")
                # keep box in memory for later...
                channelSelectorBoxIdx = box.find('Identifier').text
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Channel List":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = chanFreqPairs[0][0]
                            print("            with " + xmlVal.text)
                            break

            if box.find('Name').text == 'Frequency Band Selector':
                print("-- FREQ SELECTION BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        # POWER SPECTRUM PIPELINE: actual "freq band selector" box
                        # CONNECTIVITY PIPELINE: channel selector rebranded as freq band selector
                        if setting.find('Name').text == "Frequencies to select" or setting.find('Name').text == "Channel List":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = chanFreqPairs[0][1]
                            print("            with " + xmlVal.text)
                            break

    if len(chanFreqPairs) > 1:
        # TOUGHEST CASE : NEED TO COPY/PASTE MULTIPLE PROCESSING
        # BRANCHES WITH CHANNEL SELECTOR, TIME EPOCH, AR COEFFS,
        # FREQ BAND SELECTOR AND SPECTRUM AVERAGE

        # First box in the chain: "SPLIT" IDENTITY BOX
        splitBox = None
        splitBoxId = None
        for boxes in root.findall('Boxes'):
            for box in boxes.findall('Box'):
                if box.find('Name').text == 'SPLIT':
                    print("-- SPLIT BOX")
                    splitBox = box
                    splitBoxId = splitBox.find("Identifier").text
                    continue

        # Last box in the chain: Classifier processor
        boxLast = findBox(root, "Classifier processor")
        boxLastId = boxLast.find("Identifier").text

        featAggInputIdx = 0

        # Find all chained box btw those two
        if splitBoxId is not None and boxLastId is not None:
            locOffset = 120
            boxList = findChainedBoxes(root, splitBoxId, boxLastId)

            for idxPair, [chan, freq] in enumerate(chanFreqPairs):
                # Don't do it for the first pair, it was done earlier in the function
                if idxPair == 0:
                    continue

                pair = [chan, freq]
                # Copy list of chained boxes (except first (identity)
                # and 2 last (feature aggreg, classifier trainer)) and chain them.
                listofBoxesToChain, nbOfOutputs = copyBoxList(root, boxList, locOffset, pair, 1, 2)
                locOffset += locOffset

                # Add an input to Feature Aggregator box in the current chain
                addInputToBox(root, listofBoxesToChain[-1])
                featAggInputIdx += 1
                linkBoxes(root, listofBoxesToChain, nbOfOutputs, featAggInputIdx)

    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyConnectivityMetric(metric, scenXml):
    print("---Modifying " + scenXml + " with connectivity Metric: " + str(metric))
    tree = ET.parse(scenXml)
    root = tree.getroot()

    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == 'Connectivity Measure : Burg' or box.find('Name').text == 'Connectivity Measure':
                print("-- CONNECTIVITY BOX ")
                for settings in box.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Metric":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            xmlVal.text = str(metric)
                            print("            with " + xmlVal.text)
                            break

    # WRITE NEW XML
    tree.write(scenXml)
    return


def modifyTrainingFirstStep(runBasename, nbFeats, analysisPath, trainingPath, currentExtractId, scenXml):
    print("---Modifying " + scenXml + " outputs")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    for settings in root.findall('Settings'):
        for setting in settings.findall('Setting'):
            if setting.find('Name').text == "InputClass1":
                xmlVal = setting.find('Value')
                newinput1 = str(runBasename + "-CONNECT-REST.csv")
                xmlVal.text = newinput1
                continue
            elif setting.find('Name').text == "InputClass2":
                xmlVal = setting.find('Value')
                newinput2 = str(runBasename + "-CONNECT-MI.csv")
                xmlVal.text = newinput2
                continue
            elif setting.find('Name').text == "ExtractionIdx":
                xmlVal = setting.find('Value')
                xmlVal.text = currentExtractId
                continue

    for feat in range(nbFeats):

        for boxes in root.findall('Boxes'):
            for box in boxes.findall('Box'):
                if box.find('Name').text == str('CSV Class1 Feat'+str(feat+1)):
                    for settings in box.findall('Settings'):
                        for setting in settings.findall('Setting'):
                            if setting.find('Name').text == "Filename":
                                xmlVal = setting.find('Value')
                                print("       replacing " + xmlVal.text)
                                class1featX = os.path.join(trainingPath, currentExtractId, str(runBasename + "-class1-feat" + str(feat + 1)) + ".csv")
                                class1featX = class1featX.replace("\\", "/")
                                xmlVal.text = str(class1featX)
                                print("            with " + xmlVal.text)
                                break
                elif box.find('Name').text == str('CSV Class2 Feat'+str(feat+1)):
                    for settings in box.findall('Settings'):
                        for setting in settings.findall('Setting'):
                            if setting.find('Name').text == "Filename":
                                xmlVal = setting.find('Value')
                                print("       replacing " + xmlVal.text)
                                class2featX = os.path.join(trainingPath, currentExtractId, str(runBasename + "-class2-feat" + str(feat + 1)) + ".csv" )
                                class2featX = class2featX.replace("\\", "/")
                                xmlVal.text = str(class2featX)
                                print("            with " + xmlVal.text)
                                break
    # WRITE NEW XML
    tree.write(scenXml)
    return

def modifyTrainingSecondStep(compositeFiles, nbFeats, newWeightsName, currentExtractId, scenXml):
    print("---Modifying " + scenXml + " outputs")
    tree = ET.parse(scenXml)
    root = tree.getroot()

    # PARSE SCENARIO GENERAL SETTINGS, AND MODIFY THEM ACCORDING TO THE PROVIDED
    # PARAMETER LIST
    for settings in root.findall('Settings'):
        for setting in settings.findall('Setting'):
            if setting.find('Name').text == "OutputWeights":
                xmlVal = setting.find('Value')
                xmlVal.text = newWeightsName
                continue
            if setting.find('Name').text == "ExtractionIdx":
                xmlVal = setting.find('Value')
                xmlVal.text = currentExtractId

    featAgg = []
    featAgg.append(findBox(root, "FeatAgg1") )
    featAgg.append(findBox(root, "FeatAgg2") )
    csvBoxes = []
    csvBoxes.append(findBox(root, "CSV class1 feat1") )
    csvBoxes.append(findBox(root, "CSV class2 feat1") )

    for classIdx in [0, 1]:
        locOffset = 120

        for feat in range(nbFeats):
            # first feat : don't copy, just edit the filename
            if feat == 0:
                for settings in csvBoxes[classIdx].findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Filename":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            for file in compositeFiles:
                                if str("class" + str(classIdx + 1) + "-feat" + str(
                                        feat + 1) + "-TRAINCOMPOSITE.csv") in file:
                                    classXfeatX = file.replace("\\", "/")
                                    xmlVal.text = str(classXfeatX)
                                    print("            with " + xmlVal.text)
                                    break
                continue
            else:
                newCsvIn = copyBox(root, csvBoxes[classIdx], locOffset)
                locOffset += locOffset
                oldName = csvBoxes[classIdx].find("Name").text
                changeBoxName(newCsvIn, oldName.replace("feat1", str('feat' + str(feat+1))) )
                for settings in newCsvIn.findall('Settings'):
                    for setting in settings.findall('Setting'):
                        if setting.find('Name').text == "Filename":
                            xmlVal = setting.find('Value')
                            print("       replacing " + xmlVal.text)
                            for file in compositeFiles:
                                if str("class" + str(classIdx+1) + "-feat" + str(feat+1) + "-TRAINCOMPOSITE.csv") in file:
                                    classXfeatX = file.replace("\\", "/")
                                    xmlVal.text = str(classXfeatX)
                                    print("            with " + xmlVal.text)
                                    break

                addInputToBox(root, featAgg[classIdx].find('Identifier').text)
                linkTwoBoxesGeneric(root, newCsvIn.find("Identifier").text,
                                    featAgg[classIdx].find("Identifier").text, 0, feat)



    # WRITE NEW XML
    tree.write(scenXml)
    return

##################
# HELPER FUNCTIONS
##################

def findPreviousBox(root, boxIdx):
    for links in root.findall('Links'):
        for link in links.findall('Link'):
            targetBox = link.find("Target")
            if targetBox.find('BoxIdentifier').text == boxIdx:
                sourceBox = link.find("Source")
                previousBox = findBoxId(root, sourceBox.find('BoxIdentifier').text)
                return previousBox

    return None

def addInputToBox(root, featAggBoxIdx):
    box = findBoxId(root, featAggBoxIdx)
    inputs = box.find('Inputs')
    input = inputs.find('Input')
    newinput = copy.deepcopy(input)
    inputs.append(newinput)

def countBoxInputs(root, boxIdx):
    nbInputs = 0
    box = findBoxId(root, boxIdx)
    inputs = box.find('Inputs')
    return len(inputs.findall('Input'))

def linkBoxes(root, listofBoxesToChain, nbOfOutputs, featAggInputIdx):
    links = root.find("Links")
    for idx, boxId in enumerate(listofBoxesToChain):

        if boxId is not listofBoxesToChain[-1]:

            link = ET.SubElement(links, "Link")
            identifier = ET.SubElement(link, "Identifier")
            newId1 = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
            newId2 = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
            identifier.text = str("(0x" + str(newId1) + ", 0x" + str(newId2) + ")")

            source = ET.SubElement(link, "Source")
            sourceBoxId = ET.SubElement(source, "BoxIdentifier")
            sourceBoxId.text = str(boxId)
            sourceBoxOut = ET.SubElement(source, "BoxOutputIndex")
            # We take the last output id of the box... (works for AR coeffs, for now...)
            # WARNING : this is a baaad patch
            sourceBoxOut.text = str(nbOfOutputs[idx] - 1)
            target = ET.SubElement(link, "Target")
            targetBoxId = ET.SubElement(target, "BoxIdentifier")
            targetBoxId.text = str(listofBoxesToChain[idx+1])
            targetBoxIn = ET.SubElement(target, "BoxInputIndex")
            if idx < len(listofBoxesToChain) - 2:
                targetBoxIn.text = str(0)
            else:
                targetBoxIn.text = str(featAggInputIdx)

def linkBoxesGeneric(root, listofBoxesToChain, nbOfOutputs):
    links = root.find("Links")
    for idx, boxId in enumerate(listofBoxesToChain):

        if boxId is not listofBoxesToChain[-1]:

            link = ET.SubElement(links, "Link")
            identifier = ET.SubElement(link, "Identifier")
            newId1 = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
            newId2 = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
            identifier.text = str("(0x" + str(newId1) + ", 0x" + str(newId2) + ")")

            source = ET.SubElement(link, "Source")
            sourceBoxId = ET.SubElement(source, "BoxIdentifier")
            sourceBoxId.text = str(boxId)
            sourceBoxOut = ET.SubElement(source, "BoxOutputIndex")
            # We take the last output id of the box... (works for AR coeffs, for now...)
            # WARNING : this is a baaad patch
            sourceBoxOut.text = str(nbOfOutputs[idx] - 1)
            target = ET.SubElement(link, "Target")
            targetBoxId = ET.SubElement(target, "BoxIdentifier")
            targetBoxId.text = str(listofBoxesToChain[idx+1])
            targetBoxIn = ET.SubElement(target, "BoxInputIndex")
            targetBoxIn.text = str(0)

def linkTwoBoxesGeneric(root, boxId1, boxId2, outputidx, inputidx):
    links = root.find("Links")
    link = ET.SubElement(links, "Link")

    identifier = ET.SubElement(link, "Identifier")
    newId1 = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
    newId2 = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
    identifier.text = str("(0x" + str(newId1) + ", 0x" + str(newId2) + ")")

    source = ET.SubElement(link, "Source")
    sourceBoxId = ET.SubElement(source, "BoxIdentifier")
    sourceBoxId.text = str(boxId1)
    sourceBoxOut = ET.SubElement(source, "BoxOutputIndex")
    sourceBoxOut.text = str(outputidx)

    target = ET.SubElement(link, "Target")
    targetBoxId = ET.SubElement(target, "BoxIdentifier")
    targetBoxId.text = str(boxId2)
    targetBoxIn = ET.SubElement(target, "BoxInputIndex")
    targetBoxIn.text = str(inputidx)

def generateRandomHexId(length):
    return binascii.b2a_hex(os.urandom(length))


# Find a box by its name. Use ONLY to find a specific box in the whole tree.
# Most of times it's better to do it manually in a loop to process the xml on the fly...
def findBox(root, name):
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Name').text == name:
                return box

    return None

# Find a box by its ID. Use ONLY to find a specific box in the whole tree.
# Most of times it's better to do it manually in a loop to process the xml on the fly...
def findBoxId(root, idx):
    for boxes in root.findall('Boxes'):
        for box in boxes.findall('Box'):
            if box.find('Identifier').text == idx:
                return box

    return None

def copyBox(root, box, locOffset):
    horLocId = "(0x207c9054, 0x3c841b63)"

    boxes = root.find('Boxes')
    newbox = copy.deepcopy(box)
    # Modify some box parameters
    for node in newbox:
        # Create random id
        if node.tag == "Identifier":
            oldText = node.text
            oldIds = oldText.split(",")
            newId = str(generateRandomHexId(4)).replace("b", "").replace("\'", '')
            node.text = str(oldIds[0] + ", 0x" + str(newId) + ")")
            print("OLD ID : " + oldText + " // NEW ID : " + node.text)

        # Change location on designer
        elif node.tag == "Attributes":
            for attrib in node:
                if attrib.find("Identifier").text == horLocId:
                    loc = attrib.find('Value')
                    locInt = int(loc.text)
                    newLocInt = locInt + locOffset
                    loc.text = str(newLocInt)
                    continue

    boxes.append(newbox)
    return newbox

def copyBoxList(root, boxIdList, locOffset, chanFreqPair, nbDiscardTop, nbDiscardBottom):

    boxes = root.find('Boxes')
    # newBoxIdList & nbOfOutputs will serve for chaining later on.
    # Hence we add the first box to both lists (Stim based epoching)
    nbOfOutputs = [1]
    newBoxIdList = [boxIdList[0]]

    # List to parse : discard boxes we don't want to copy
    boxIdListToParse = boxIdList.copy()
    for i in range(nbDiscardTop):
        boxIdListToParse.pop(0)
    for i in range(nbDiscardBottom):
        boxIdListToParse.pop()

    for boxId in boxIdListToParse:
        # Find actual box in root and copy it
        for box in boxes:
            if box.find('Identifier').text == boxId:
                newBox = copyBox(root, box, locOffset)
                newBoxId = newBox.find('Identifier').text
                newBoxIdList.append(newBoxId)
                outputs = box.find('Outputs')
                nbOfOutputs.append(0)
                for output in outputs:
                    nbOfOutputs[-1] += 1

                # EDIT CHANNEL SELECTOR AND FREQ SELECTOR PARAMETERS
                if newBox.find('Name').text == "Channel Selector":
                    for settings in box.findall('Settings'):
                        for setting in settings.findall('Setting'):
                            if setting.find('Name').text == "Channel List":
                                xmlVal = setting.find('Value')
                                print("       replacing " + xmlVal.text)
                                xmlVal.text = chanFreqPair[0]
                                print("            with " + xmlVal.text)
                                continue
                elif newBox.find('Name').text == "Frequency Band Selector":
                    for settings in box.findall('Settings'):
                        for setting in settings.findall('Setting'):
                            # POWER SPECTRUM PIPELINE: actual "freq band selector" box
                            # CONNECTIVITY PIPELINE: channel selector rebranded as freq band selector
                            if setting.find('Name').text == "Frequencies to select" or setting.find('Name').text == "Channel List":
                                xmlVal = setting.find('Value')
                                print("       replacing " + xmlVal.text)
                                xmlVal.text = chanFreqPair[1]
                                print("            with " + xmlVal.text)
                                continue

                break

    # Add last element of the list, for future linking/chaining
    if len(newBoxIdList) > 1:
        newBoxIdList.append(boxIdList[len(boxIdList)-2])
        nbOfOutputs.append(1)

    print(newBoxIdList)

    return newBoxIdList, nbOfOutputs


def copyBoxListGeneric(root, boxIdList, locOffset, chanFreqPair, nbDiscardTop, nbDiscardBottom):

    boxes = root.find('Boxes')
    # newBoxIdList & nbOfOutputs will serve for chaining later on.
    # Hence we add the first box to both lists (Stim based epoching)
    nbOfOutputs = [1]
    newBoxIdList = [boxIdList[0]]

    # List to parse : discard boxes we don't want to copy
    boxIdListToParse = boxIdList.copy()
    for i in range(nbDiscardTop):
        boxIdListToParse.pop(0)
    for i in range(nbDiscardBottom):
        boxIdListToParse.pop()

    for boxId in boxIdListToParse:
        # Find actual box in root and copy it
        for box in boxes:
            if box.find('Identifier').text == boxId:
                newBox = copyBox(root, box, locOffset)
                newBoxId = newBox.find('Identifier').text
                newBoxIdList.append(newBoxId)
                outputs = box.find('Outputs')
                nbOfOutputs.append(0)
                if outputs:
                    for output in outputs:
                        nbOfOutputs[-1] += 1

                # EDIT CHANNEL SELECTOR AND FREQ SELECTOR PARAMETERS
                if newBox.find('Name').text == "Channel Selector":
                    for settings in box.findall('Settings'):
                        for setting in settings.findall('Setting'):
                            if setting.find('Name').text == "Channel List":
                                xmlVal = setting.find('Value')
                                print("       replacing " + xmlVal.text)
                                xmlVal.text = chanFreqPair[0]
                                print("            with " + xmlVal.text)
                                continue
                elif newBox.find('Name').text == "Frequency Band Selector":
                    for settings in box.findall('Settings'):
                        for setting in settings.findall('Setting'):
                            # POWER SPECTRUM PIPELINE: actual "freq band selector" box
                            # CONNECTIVITY PIPELINE: channel selector rebranded as freq band selector
                            if setting.find('Name').text == "Frequencies to select" or setting.find('Name').text == "Channel List":
                                xmlVal = setting.find('Value')
                                print("       replacing " + xmlVal.text)
                                xmlVal.text = chanFreqPair[1]
                                print("            with " + xmlVal.text)
                                continue

                break

    # For future linking/chaining
    if len(newBoxIdList) > 1:
        nbOfOutputs.append(1)

    print(newBoxIdList)

    return newBoxIdList, nbOfOutputs


# Find (linear) chain of boxes. No branches please !
def findChainedBoxes(root, firstBoxId, lastBoxId):
    boxList = [firstBoxId]
    condition = True

    foundLast = False

    for links in root.findall('Links'):

        while condition:
            found = False

            for link in links.findall('Link'):
                sourceBox = link.find("Source")
                id = sourceBox.find('BoxIdentifier')
                if sourceBox.find('BoxIdentifier').text == boxList[-1]:
                    targetBox = link.find("Target")
                    boxList.append(targetBox.find('BoxIdentifier').text)
                    found = True
                    if boxList[-1] == lastBoxId:
                        foundLast = True
                    break

            if not found or foundLast:
                condition = False

    if not foundLast:
        boxList = []

    print(boxList)
    return boxList

def changeBoxName(box, newName):
    box.find('Name').text = newName
    return




import sys
import os
import time
import subprocess
import platform
from shutil import copyfile
from importlib import resources

from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal

from happyfeat.lib.mergeRunsCsv import mergeRunsCsv, mergeRunsCsv_new
from happyfeat.lib.extractMetaData import extractMetadata, generateMetadata
from happyfeat.lib.modifyOpenvibeScen import *
from happyfeat.lib.Visualization_Data import *
from happyfeat.lib.featureExtractUtils import *
from happyfeat.lib.utils import *

from happyfeat.lib.bcipipeline_settings import *

# ------------------------------------------------------
# CLASSES FOR LONG-RUNNING OPERATIONS IN THREADS
# ------------------------------------------------------

class Acquisition(QtCore.QThread):
    over = pyqtSignal(bool, str)

    def __init__(self, ovScript, scenFile, parameterDict, parent=None):
        super().__init__(parent)
        self.stop = False
        self.ovScript = ovScript
        self.scenFile = scenFile
        self.parameterDict = parameterDict.copy()

    def run(self):
        command = self.ovScript
        if platform.system() == 'Windows':
            command = command.replace("/", "\\")

        exitText = ""
        success = True
        p = subprocess.Popen([command, "--no-gui", "--play", self.scenFile],
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        # Print console output, and detect end of process...
        while True:
            output = p.stdout.readline()
            if p.poll() is not None:
                break
            if output:
                print(str(output))
                if "Application terminated" in str(output):
                    break
                if "Could not connect to server" in str(output):
                    exitText = str("Error running \"Acquisition\" scenario\n")
                    exitText += str("Make sure the Acquisition Server is running and in \"Play\" mode\n")
                    success = False
                    break

        self.stop = True
        self.over.emit(success, exitText)

    def stopThread(self):
        self.stop = True

class Extraction(QtCore.QThread):
    info = pyqtSignal(bool)
    info2 = pyqtSignal(str)
    over = pyqtSignal(bool, str)

    def __init__(self, ovScript, scenFile, signalFiles, signalFolder,
                 parameterDict, currentSessionId, parent=None):

        super().__init__(parent)
        self.stop = False
        self.ovScript = ovScript
        self.scenFile = scenFile
        self.signalFiles = signalFiles
        self.signalFolder = signalFolder
        self.parameterDict = parameterDict.copy()
        self.currentSessionId = currentSessionId
        self.extractDict = parameterDict["Sessions"][currentSessionId]["ExtractionParams"].copy()

    def run(self):
        command = self.ovScript
        if platform.system() == 'Windows':
            command = command.replace("/", "\\")

        for signalFile in self.signalFiles:
            tstart = time.perf_counter()

            self.info2.emit(str("Extracting data for file " + signalFile + "..."))

            # Verify the existence of metadata files for each selected files,
            # and if not, generate them.
            # Then extract sampling frequency and electrode list
            sampFreq = None
            electrodeList = None
            metaFile = signalFile.replace(".ov", "-META.csv")

            if metaFile in os.listdir(self.signalFolder):
                sampFreq, electrodeList = extractMetadata(os.path.join(self.signalFolder, metaFile))
            else:
                generateMetadata(os.path.join(self.signalFolder, signalFile), self.ovScript)
                sampFreq, electrodeList = extractMetadata(os.path.join(self.signalFolder, metaFile))
            # Check everything went ok...
            if not sampFreq:
                errMsg = str("Error while loading metadata CSV file for session " + signalFile)
                self.over.emit(False, errMsg)
                return

            ## MODIFY THE EXTRACTION SCENARIO with entered parameters
            # /!\ after updating ARburg order and FFT size using sampfreq
            self.extractDict["ChannelNames"] = ";".join(electrodeList)
            self.extractDict["AutoRegressiveOrder"] = str(
                timeToSamples(float(self.extractDict["AutoRegressiveOrderTime"]), sampFreq))
            self.extractDict["PsdSize"] = str(freqResToPsdSize(float(self.extractDict["FreqRes"]), sampFreq))

            # Special case : "subset of electrodes" for connectivity
            # DISABLED FOR NOW
            # if field is empty, use all electrodes
            # if self.parameterDict["pipelineType"] == settings.optionKeys[2]:
            #     if self.parameterDict["ChannelSubset"] == "":
            #         self.parameterDict["ChannelSubset"] = self.parameterDict["ChannelNames"]
            self.extractDict["ChannelSubset"] = self.extractDict["ChannelNames"]

            # Special case : "connectivity shift", used in scenarios but not set that way
            # in the interface
            if "ConnectivityOverlap" in self.extractDict.keys():
                self.extractDict["ConnectivityShift"] = str(float(self.extractDict["ConnectivityLength"]) * (100.0 - float(self.extractDict["ConnectivityOverlap"])) / 100.0)

            # Modify the scenario
            modifyScenarioGeneralSettings(self.scenFile, self.extractDict)

            # Special case: "connectivity metric"
            if "ConnectivityMetric" in self.extractDict.keys():
                modifyConnectivityMetric(self.extractDict["ConnectivityMetric"], self.scenFile)

            # Modify extraction scenario to use provided signal file, and rename outputs accordingly
            filename = signalFile.removesuffix(".ov")
            outputSpect1 = str(filename + "-SPECTRUM-" + self.parameterDict["AcquisitionParams"]["Class1"] + ".csv")
            outputSpect2 = str(filename + "-SPECTRUM-" + self.parameterDict["AcquisitionParams"]["Class2"] + ".csv")
            outputConnect1 = str(filename + "-CONNECT-" + self.parameterDict["AcquisitionParams"]["Class1"] + ".csv")
            outputConnect2 = str(filename + "-CONNECT-" + self.parameterDict["AcquisitionParams"]["Class2"] + ".csv")
            outputBaseline1 = str(filename + "-SPECTRUM-" + self.parameterDict["AcquisitionParams"]["Class1"] + "-BASELINE.csv")
            outputBaseline2 = str(filename + "-SPECTRUM-" + self.parameterDict["AcquisitionParams"]["Class2"] + "-BASELINE.csv")
            outputTrials = str(filename + "-TRIALS.csv")
            modifyExtractionIO(self.scenFile, signalFile, outputSpect1, outputSpect2,
                               outputBaseline1, outputBaseline2, outputConnect1, outputConnect2, outputTrials, self.currentSessionId)

            # Launch OV scenario !
            p = subprocess.Popen([command, "--invisible", "--play-fast", self.scenFile],
                                 stdin=subprocess.PIPE, stdout=subprocess.PIPE)

            # Print console output, and detect end of process...
            while True:
                output = p.stdout.readline()
                if p.poll() is not None:
                    break
                if output:
                    print(str(output))
                    if "Application terminated" in str(output):
                        break

            self.info.emit(True)    # send "info" signal to increment progressbar
            tstop = time.perf_counter()
            print("= Extraction from file " + filename + " finished in " + str(tstop-tstart))

        self.stop = True
        self.over.emit(True, "")

    def stopThread(self):
        self.stop = True


class LoadFilesForVizPowSpectrum(QtCore.QThread):
    info = pyqtSignal(bool)
    info2 = pyqtSignal(str)
    over = pyqtSignal(bool, str)

    def __init__(self, analysisFiles, workingFolder, parameterDict, Features, sampFreq, parent=None):

        super().__init__(parent)
        self.stop = False
        self.analysisFiles = analysisFiles
        self.workingFolder = workingFolder
        self.parameterDict = parameterDict.copy()
        self.extractDict = parameterDict["Sessions"][parameterDict["currentSessionId"]]["ExtractionParams"].copy()
        self.Features = Features
        self.samplingFreq = sampFreq

        self.dataNp1 = []
        self.dataNp2 = []
        self.dataNp1baseline = []
        self.dataNp2baseline = []

    def run(self):

        listSampFreq = []
        listElectrodeList = []
        listFreqBins = []
        idxFile = 0

        for selectedFilesForViz in self.analysisFiles:
            idxFile += 1
            pipelineLabel = "SPECTRUM"
            class1label = self.parameterDict["AcquisitionParams"]["Class1"]
            class2label = self.parameterDict["AcquisitionParams"]["Class2"]
            selectedBasename = selectedFilesForViz

            path1 = os.path.join(self.workingFolder, str(selectedBasename + "-" + pipelineLabel + "-" + class1label + ".csv"))
            path2 = os.path.join(self. workingFolder, str(selectedBasename + "-" + pipelineLabel + "-" + class2label + ".csv"))
            path1baseline = os.path.join(self.workingFolder, str(selectedBasename + "-" + pipelineLabel + "-" + class1label + "-BASELINE.csv"))
            path2baseline = os.path.join(self.workingFolder, str(selectedBasename + "-" + pipelineLabel + "-" + class2label + "-BASELINE.csv"))

            self.info2.emit(str("Loading " + pipelineLabel + " Data for file " + str(idxFile) + " : " + selectedFilesForViz))
            [header1, data1] = load_csv_np(path1)
            [header2, data2] = load_csv_np(path2)
            self.info2.emit(str("Loading " + pipelineLabel + " Baseline Data for file " + str(idxFile) + " : " + selectedFilesForViz))
            [header1baseline, data1baseline] = load_csv_np(path1baseline)
            [header2baseline, data2baseline] = load_csv_np(path2baseline)

            # Sampling frequency
            # Infos in the columns header of the CSVs in format "Time:32x251:500"
            # (Column zero contains starting time of the row)
            # 32 is channels, 251 is freq bins, 500 is sampling frequency)
            sampFreq1 = int(header1[0].split(":")[-1])
            sampFreq2 = int(header2[0].split(":")[-1])
            freqBins1 = int(header1[0].split(":")[1].split("x")[1])
            freqBins2 = int(header2[0].split(":")[1].split("x")[1])
            if sampFreq1 != sampFreq2 or freqBins1 != freqBins2:
                errMsg = str("Error when loading " + path1 + "\n" + " and " + path2)
                errMsg = str(errMsg + "\nSampling frequency or frequency bins mismatch")
                errMsg = str(errMsg + "\n(" + str(sampFreq1) + " vs " + str(sampFreq2) + " or ")
                errMsg = str(errMsg + str(freqBins1) + " vs " + str(freqBins2) + ")")
                self.over.emit(False, errMsg)
                return

            listSampFreq.append(sampFreq1)
            listFreqBins.append(freqBins1)

            elecTemp1 = header1[2:-3]
            elecTemp2 = header2[2:-3]
            electrodeList1 = []
            electrodeList2 = []
            for i in range(0, len(elecTemp1), freqBins1):
                electrodeList1.append(elecTemp1[i].split(":")[0])
                electrodeList2.append(elecTemp2[i].split(":")[0])

            if electrodeList1 != electrodeList2:
                errMsg = str("Error when loading " + path1 + "\n" + " and " + path2)
                errMsg = str(errMsg + "\nElectrode List mismatch")
                self.over.emit(False, errMsg)
                return

            listElectrodeList.append(electrodeList1)

            self.dataNp1.append(data1)
            self.dataNp2.append(data2)
            self.dataNp1baseline.append(data1baseline)
            self.dataNp2baseline.append(data2baseline)
            self.info.emit(True)

        # Check if all files have the same sampling freq and electrode list. If not, for now, we don't process further
        if not all(freqsamp == listSampFreq[0] for freqsamp in listSampFreq):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Sampling frequency mismatch (" + str(listSampFreq) + ")")
            self.over.emit(False, errMsg)
            return
        else:
            self.samplingFreq = listSampFreq[0]
            print("Sampling Frequency for selected files : " + str(listSampFreq[0]))

        if not all(electrodeList == listElectrodeList[0] for electrodeList in listElectrodeList):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Electrode List mismatch")
            self.over.emit(False, errMsg)
            return
        else:
            print("Sensor list for selected files : " + ";".join(listElectrodeList[0]))

        if not all(freqBins == listFreqBins[0] for freqBins in listFreqBins):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Not same number of frequency bins (" + str(listSampFreq) + ")")
            self.over.emit(False, errMsg)
            return
        else:
            print("Frequency bins: " + str(listFreqBins[0]))

        # ----------
        # Compute the features used for visualization
        # ----------
        trialLength = float(self.extractDict["StimulationEpoch"])
        trials = int(self.parameterDict["AcquisitionParams"]["TrialNb"])
        electrodeList = listElectrodeList[0]
        nbElectrodes = len(electrodeList)
        n_bins = listFreqBins[0]
        winLen = float(self.extractDict["TimeWindowLength"])
        winShift = float(self.extractDict["TimeWindowShift"])

        # For multiple runs (ie. multiple selected CSV files), we just concatenate
        # the trials from all files. Then the displayed spectral features (R²map, PSD, topography)
        # will be computed as averages over all the trials.
        power_cond1_final = None
        power_cond2_final = None
        power_cond1_baseline_final = None
        power_cond2_baseline_final = None
        timefreq_cond1_final = None
        timefreq_cond2_final = None
        timefreq_cond1_baseline_final = None
        timefreq_cond2_baseline_final = None
        idxFile = 0
        for run in range(len(self.dataNp1)):
            idxFile += 1
            self.info2.emit(str("Processing data for file " + str(idxFile)))
            power_cond1, timefreq_cond1 = \
                Extract_CSV_Data(self.dataNp1[run], trialLength, nbElectrodes, n_bins, winLen, winShift)
            power_cond2, timefreq_cond2 = \
                Extract_CSV_Data(self.dataNp2[run], trialLength, nbElectrodes, n_bins, winLen, winShift)
            power_cond1_baseline, timefreq_cond1_baseline = \
                Extract_CSV_Data(self.dataNp1baseline[run], trialLength, nbElectrodes, n_bins, winLen, winShift)
            power_cond2_baseline, timefreq_cond2_baseline = \
                Extract_CSV_Data(self.dataNp2baseline[run], trialLength, nbElectrodes, n_bins, winLen, winShift)

            if power_cond1_final is None:
                power_cond1_final = power_cond1
                power_cond2_final = power_cond2
                power_cond1_baseline_final = power_cond1_baseline
                power_cond2_baseline_final = power_cond2_baseline
                timefreq_cond1_final = timefreq_cond1
                timefreq_cond2_final = timefreq_cond2
                timefreq_cond1_baseline_final = timefreq_cond1_baseline
                timefreq_cond2_baseline_final = timefreq_cond2_baseline
            else:
                power_cond1_final = np.concatenate((power_cond1_final, power_cond1))
                power_cond2_final = np.concatenate((power_cond2_final, power_cond2))
                power_cond1_baseline_final = np.concatenate((power_cond1_baseline_final, power_cond1_baseline))
                power_cond2_baseline_final = np.concatenate((power_cond2_baseline_final, power_cond2_baseline))
                timefreq_cond1_final = np.concatenate((timefreq_cond1_final, timefreq_cond1))
                timefreq_cond2_final = np.concatenate((timefreq_cond2_final, timefreq_cond2))
                timefreq_cond1_baseline_final = np.concatenate((timefreq_cond1_baseline_final, timefreq_cond1_baseline))
                timefreq_cond2_baseline_final = np.concatenate((timefreq_cond2_baseline_final, timefreq_cond2_baseline))

        self.info2.emit("Computing statistics")

        trialLengthSec = float(self.parameterDict["AcquisitionParams"]["TrialLength"])
        totalTrials = len(self.dataNp1) * trials
        windowLength = float(self.extractDict["TimeWindowLength"])
        windowShift = float(self.extractDict["TimeWindowShift"])
        segmentsPerTrial = round((trialLength - windowLength) / windowShift)
        fres = float(self.extractDict["FreqRes"])

        timeVectAtomic = [0]
        for i in range(segmentsPerTrial - 1):
            timeVectAtomic.append((i + 1) * windowShift)

        timeVectAtomic = np.array(timeVectAtomic)
        time_array = np.empty(0)
        idxTrial = 0
        for trial in range(totalTrials):
            time_array = np.concatenate((time_array, timeVectAtomic + (idxTrial * trialLengthSec)))
            idxTrial += 1

        # Statistical Analysis
        freqs_array = np.arange(0, n_bins, fres)

        Rsigned = Compute_Rsquare_Map(power_cond2_final[:, :, :(n_bins - 1)],
                                      power_cond1_final[:, :, :(n_bins - 1)])
        Wsquare, Wpvalues = Compute_Wilcoxon_Map(power_cond2_final[:, :, :(n_bins - 1)],
                                                 power_cond1_final[:, :, :(n_bins - 1)])

        # Reordering for R map and topography...
        if self.parameterDict["sensorMontage"] == "standard_1020" \
            or self.parameterDict["sensorMontage"] == "biosemi64":
            Rsigned_2, Wsquare_2, Wpvalues_2, electrodes_final, power_cond1_2, power_cond2_2, timefreq_cond1_2, timefreq_cond2_2 \
                = Reorder_plusplus(Rsigned, Wsquare, Wpvalues, electrodeList, power_cond1_final, power_cond2_final,
                                   timefreq_cond1_final, timefreq_cond2_final)
        elif self.parameterDict["sensorMontage"] == "custom" \
            and self.parameterDict["customMontagePath"] != "":
            Rsigned_2, Wsquare_2, Wpvalues_2, electrodes_final, power_cond1_2, power_cond2_2, timefreq_cond1_2, timefreq_cond2_2 \
                = Reorder_custom_plus(Rsigned, Wsquare, Wpvalues, self.parameterDict["customMontagePath"], electrodeList, power_cond1_final, power_cond2_final,
                                   timefreq_cond1_final, timefreq_cond2_final)

        self.Features.electrodes_orig = electrodeList
        self.Features.power_cond2 = power_cond2_2
        self.Features.power_cond1 = power_cond1_2
        self.Features.timefreq_cond1 = timefreq_cond1_2
        self.Features.timefreq_cond2 = timefreq_cond2_2
        # self.Features.time_array = time_array
        self.Features.time_array = timeVectAtomic
        self.Features.freqs_array = freqs_array
        self.Features.fres = fres
        self.Features.electrodes_final = electrodes_final
        self.Features.Rsigned = Rsigned_2
        self.Features.Wsigned = Wsquare_2

        self.Features.average_baseline_cond1 = np.mean(power_cond1_baseline_final, axis=0)
        self.Features.std_baseline_cond1 = np.std(power_cond1_baseline_final, axis=0)
        self.Features.average_baseline_cond2 = np.mean(power_cond2_baseline_final, axis=0)
        self.Features.std_baseline_cond2 = np.std(power_cond2_baseline_final, axis=0)

        self.Features.samplingFreq = self.samplingFreq

        self.stop = True
        self.over.emit(True, "")

    def stopThread(self):
        self.stop = True

# Class for loading extracted data in the "Connectivity Pipeline"
# data are NODE STRENGTH (not raw connectivity matrices), we manage them
# almost like power spectra...
class LoadFilesForVizConnectivity(QtCore.QThread):
    info = pyqtSignal(bool)
    info2 = pyqtSignal(str)
    over = pyqtSignal(bool, str)

    def __init__(self, analysisFiles, workingFolder, metaFolder, parameterDict, Features, sampFreq, parent=None):

        super().__init__(parent)
        self.stop = False
        self.analysisFiles = analysisFiles
        self.workingFolder = workingFolder
        self.metaFolder = metaFolder
        self.parameterDict = parameterDict.copy()
        self.extractDict = parameterDict["Sessions"][parameterDict["currentSessionId"]]["ExtractionParams"].copy()
        self.Features = Features
        self.samplingFreq = sampFreq

        self.dataNp1 = []
        self.dataNp2 = []

    def run(self):

        listSamplingFreqs = []
        listFreqs = []
        listElectrodeList = []
        idxFile = 0

        for selectedFilesForViz in self.analysisFiles:
            idxFile += 1
            pipelineLabel = "CONNECT"
            class1label = self.parameterDict["AcquisitionParams"]["Class1"]
            class2label = self.parameterDict["AcquisitionParams"]["Class2"]
            selectedBasename = selectedFilesForViz

            # First, get metadata...
            metapath = os.path.join(self.metaFolder, str(selectedBasename + "-META.csv"))
            [sampFreq, electrodeList] = extractMetadata(metapath)
            listSamplingFreqs.append(sampFreq)

            # Now load data...
            path1 = os.path.join(self.workingFolder, str(selectedBasename + "-" + pipelineLabel + "-" + class1label + ".csv"))
            path2 = os.path.join(self.workingFolder, str(selectedBasename + "-" + pipelineLabel + "-" + class2label + ".csv"))

            self.info2.emit(str("Loading " + pipelineLabel + " Data for file " + str(idxFile) + " : " + selectedFilesForViz))
            [header1, data1] = load_csv_np(path1)
            [header2, data2] = load_csv_np(path2)

            # Check that files match...!
            # Infos in the columns header of the CSVs in format "Time:500x64"
            # (Column zero contains starting time of the row)
            # 500 is nb of freq bins, 64 is channels)
            freqBins1 = int(header1[0].split(":")[-1].split("x")[0])
            freqBins2 = int(header2[0].split(":")[-1].split("x")[0])
            if freqBins1 != freqBins2:
                errMsg = str("Error when loading " + path1 + "\n" + " and " + path2)
                errMsg = str(errMsg + "\nfrequency bins mismatch")
                errMsg = str(errMsg + "\n(" + str(freqBins1) + " vs " + str(freqBins2) + ")")
                self.over.emit(False, errMsg)
                return

            listFreqs.append(freqBins1)

            # check match between electrodes...
            nbElec1 = int(header1[0].split(":")[-1].split("x")[1])
            nbElec2 = int(header2[0].split(":")[-1].split("x")[1])
            elecTemp1 = header1[2:nbElec1+2]
            elecTemp2 = header2[2:nbElec2+2]
            electrodeList1 = []
            electrodeList2 = []
            for i in range(0, len(elecTemp1)):
                electrodeList1.append(elecTemp1[i].split(":")[1])
                electrodeList2.append(elecTemp2[i].split(":")[1])

            if electrodeList1 != electrodeList2:
                errMsg = str("Error when loading " + path1 + "\n" + " and " + path2)
                errMsg = str(errMsg + "\nElectrode List mismatch")
                self.over.emit(False, errMsg)
                return

            listElectrodeList.append(electrodeList1)

            self.dataNp1.append(data1)
            self.dataNp2.append(data2)

            self.info.emit(True)

        # Check if all files have the same sampling freq and electrode list. If not, for now, we don't process further
        if not all(nbfreqs == listFreqs[0] for nbfreqs in listFreqs):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "nb of frequency mismatch (" + str(listFreqs) + ")")
            self.over.emit(False, errMsg)
            return
        else:
            print("Nb of Frequency bins for selected files : " + str(listFreqs[0]))

        if not all(electrodeList == listElectrodeList[0] for electrodeList in listElectrodeList):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Sensor List mismatch")
            self.over.emit(False, errMsg)
            return
        else:
            print("Sensor list for selected files : " + ";".join(listElectrodeList[0]))

        if not all(sampFreq == listSamplingFreqs[0] for sampFreq in listSamplingFreqs):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Sampling Freq mismatch")
            self.over.emit(False, errMsg)
            return
        else:
            print("Sensor list for selected files : " + ";".join(listElectrodeList[0]))

        # ----------
        # Compute the features used for visualization
        # ----------
        trialLength = float(self.extractDict["StimulationEpoch"])
        trials = int(self.parameterDict["AcquisitionParams"]["TrialNb"])
        electrodeList = listElectrodeList[0]
        nbElectrodes = len(electrodeList)
        n_bins = listFreqs[0]
        connectLength = float(self.extractDict["ConnectivityLength"])
        connectOverlap = float(self.extractDict["ConnectivityOverlap"])
        sampFreq = listSamplingFreqs[0]

        # For multiple runs (ie. multiple selected CSV files), we just concatenate
        # the trials from all files. Then the displayed spectral features (R²map, PSD, topography)
        # will be computed as averages over all the trials.
        connect_cond1_final = None
        connect_cond2_final = None
        idxFile = 0
        for run in range(len(self.dataNp1)):
            idxFile += 1
            self.info2.emit(str("Processing data for file " + str(idxFile)))
            connect_cond1 = Extract_Connect_NodeStrength_CSV_Data(self.dataNp1[run], trialLength, nbElectrodes, n_bins,
                                                                  connectLength, connectOverlap)
            connect_cond2 = Extract_Connect_NodeStrength_CSV_Data(self.dataNp2[run], trialLength, nbElectrodes, n_bins,
                                                                  connectLength, connectOverlap)
            if connect_cond1_final is None:
                connect_cond1_final = connect_cond1
                connect_cond2_final = connect_cond2
            else:
                connect_cond1_final = np.concatenate((connect_cond1_final, connect_cond1))
                connect_cond2_final = np.concatenate((connect_cond2_final, connect_cond2))

        self.info2.emit("Computing statistics")
        trialLengthSec = float(self.parameterDict["AcquisitionParams"]["TrialLength"])
        totalTrials = len(self.dataNp1) * trials
        fres = float(self.extractDict["FreqRes"])

        # Statistical Analysis...
        freqs_array = np.arange(0, n_bins, fres)
        Rsigned = Compute_Rsquare_Map(connect_cond2_final[:, :, :(n_bins - 1)],
                                      connect_cond1_final[:, :, :(n_bins - 1)])


        # Reordering for R map and topography...
        if self.parameterDict["sensorMontage"] == "standard_1020" \
            or self.parameterDict["sensorMontage"] == "biosemi64":
            Rsigned_2, electrodes_final, connect_cond1_2, connect_cond2_2, \
                = Reorder_Rsquare(Rsigned, electrodeList, connect_cond1_final, connect_cond2_final)

        elif self.parameterDict["sensorMontage"] == "custom" \
            and self.parameterDict["customMontagePath"] != "":
            Rsigned_2, electrodes_final, connect_cond1_2, connect_cond2_2, \
                = Reorder_custom(Rsigned, self.parameterDict["customMontagePath"], electrodeList, connect_cond1_final, connect_cond2_final,)

        # Fill Features struct...
        self.Features.electrodes_orig = electrodeList
        self.Features.electrodes_final = electrodes_final
        self.Features.power_cond1 = connect_cond1_2
        self.Features.power_cond2 = connect_cond2_2
        self.Features.fres = fres
        self.Features.samplingFreq = sampFreq
        self.Features.freqs_array = freqs_array

        self.Features.Rsigned = Rsigned_2

        self.stop = True
        self.over.emit(True, "")

    def stopThread(self):
        self.stop = True


class TrainClassifier(QtCore.QThread):
    info = pyqtSignal(bool)
    info2 = pyqtSignal(str)
    over = pyqtSignal(bool, str)

    def __init__(self, trainingFiles,
                 signalFolder, templateFolder,
                 workspaceFolder, ovScript,
                 trainingSize, selectedFeats,
                 parameterDict, sampFreq, currentAttempt, attemptId,
                 speedUp, parent=None):

        super().__init__(parent)
        self.stop = False
        self.trainingFiles = trainingFiles
        self.signalFolder = signalFolder
        self.templateFolder = templateFolder
        self.workspaceFolder = workspaceFolder
        self.ovScript = ovScript
        self.trainingSize = trainingSize
        self.currentAttempt = currentAttempt
        self.attemptId = attemptId
        self.speedUp = speedUp

        # selectedFeats is either a list of feats. of interest
        # or (case of mixed features) a list of 2 lists
        self.selectedFeats = selectedFeats
        self.parameterDict = parameterDict.copy()
        self.currentSessionId = self.parameterDict["currentSessionId"]
        self.extractDict = parameterDict["Sessions"][self.parameterDict["currentSessionId"]]["ExtractionParams"].copy()
        self.samplingFreq = sampFreq
        self.exitText = ""

        self.usingDualFeatures = self.parameterDict["pipelineType"] == optionKeys[3]

    def run(self):
        # Get electrodes lists and sampling freqs, and check that they match
        # + that the selected channels are in the list of electrodes
        trainingSigList = []
        listSampFreq = []
        listElectrodeList = []
        for trainingFile in self.trainingFiles:
            path = os.path.join(self.workspaceFolder, "sessions", self.currentSessionId, "train", trainingFile)
            header = pd.read_csv(path, nrows=0).columns.tolist()
            listSampFreq.append(int(header[0].split(':')[1].removesuffix('Hz')))
            listElectrodeList.append(header[2:-3])
            trainingSigList.append(path)

        if not all(freqsamp == listSampFreq[0] for freqsamp in listSampFreq):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Sampling frequency mismatch (" + str(listSampFreq) + ")")
            self.over.emit(False, errMsg)
            return
        else:
            print("Sampling Frequency for selected files : " + str(listSampFreq[0]))

        if not all(electrodeList == listElectrodeList[0] for electrodeList in listElectrodeList):
            errMsg = str("Error when loading CSV files\n")
            errMsg = str(errMsg + "Electrode List mismatch")
            self.over.emit(False, errMsg)
            return
        else:
            print("Sensor list for selected files : " + ";".join(listElectrodeList[0]))

        selectedFeats = None
        selectedFeats2 = None
        if not self.usingDualFeatures:
            selectedFeats, errMsg = self.checkSelectedFeats(self.selectedFeats, listSampFreq[0], listElectrodeList[0])
            if not selectedFeats:
                self.over.emit(False, errMsg)
                return
        else:
            selectedFeats, errMsg = self.checkSelectedFeats(self.selectedFeats[0], listSampFreq[0], listElectrodeList[0])
            if not selectedFeats:
                self.over.emit(False, errMsg)
                return
            selectedFeats2, errMsg = self.checkSelectedFeats(self.selectedFeats[1], listSampFreq[0], listElectrodeList[0])
            if not selectedFeats2:
                self.over.emit(False, errMsg)
                return

        epochCount = [0, 0]
        stimEpochLength = float(self.extractDict["StimulationEpoch"])
        if self.parameterDict["pipelineType"] == optionKeys[1]:
            winLength = float(self.extractDict["TimeWindowLength"])
            winShift = float(self.extractDict["TimeWindowShift"])
            epochCount[0] = np.floor((stimEpochLength - winLength) / winShift) + 1
        elif self.parameterDict["pipelineType"] == optionKeys[2]:
            winLength = float(self.extractDict["ConnectivityLength"])
            overlap = float(self.extractDict["ConnectivityOverlap"])
            winShift = winLength * (100.0-overlap) / 100.0
            epochCount[0] = np.floor((stimEpochLength - winLength) / winShift) + 1
        elif self.parameterDict["pipelineType"] == optionKeys[3]:
            winLength0 = float(self.extractDict["TimeWindowLength"])
            winShift0 = float(self.extractDict["TimeWindowShift"])
            epochCount[0] = np.floor((stimEpochLength - winLength0) / winShift0) + 1
            winLength1 = float(self.extractDict["ConnectivityLength"])
            overlap = float(self.extractDict["ConnectivityOverlap"])
            winShift1 = winLength1 * (100.0-overlap) / 100.0
            epochCount[1] = np.floor((stimEpochLength - winLength1) / winShift1) + 1

        ## MODIFY THE SCENARIO with entered parameters
        # /!\ after updating ARburg order and FFT size using sampfreq
        self.extractDict["ChannelNames"] = ";".join(listElectrodeList[0])
        self.extractDict["AutoRegressiveOrder"] = str(
            timeToSamples(float(self.extractDict["AutoRegressiveOrderTime"]), listSampFreq[0]))
        self.extractDict["PsdSize"] = str(freqResToPsdSize(float(self.extractDict["FreqRes"]), listSampFreq[0]))

        # Special case : "connectivity shift", used in scenarios but not set that way
        # in the interface
        if "ConnectivityOverlap" in self.extractDict.keys():
            self.extractDict["ConnectivityShift"] = str(float(self.extractDict["ConnectivityLength"]) * (100.0 - float(self.extractDict["ConnectivityOverlap"])) / 100.0)

        # Case of a single feature type (power spectrum OR connectivity...)
        # RE-COPY sc2 & sc3 FROM TEMPLATE, SO THE USER CAN DO THIS MULTIPLE TIMES
        for i in [2, 3, 4, 5]:
            scenName = templateScenFilenames[i]
            print("---Copying file from folder " + str(__name__.split('.')[0] + '.' + self.templateFolder))
            with resources.path(str(__name__.split('.')[0] + '.' + self.templateFolder), scenName) as srcFile:
                destFile = os.path.join(self.workspaceFolder, scenName)
                print("---Copying file " + str(srcFile) + " to " + str(destFile))
                copyfile(srcFile, destFile)
            trainingpath = os.path.join(self.workspaceFolder, "sessions", self.currentSessionId, "train")
            modifyScenarioGeneralSettings(str(destFile), self.extractDict)
            if i == 2:
                # training scenarios
                if not self.usingDualFeatures:
                    modifyTrainScenUsingSplitAndClassifiers("SPLIT", "Classifier trainer", selectedFeats, epochCount[0], destFile)
                    # Special case: "connectivity metric"
                    if "ConnectivityMetric" in self.extractDict.keys():
                        modifyConnectivityMetric(self.extractDict["ConnectivityMetric"], destFile)
                else:
                    modifyTrainScenUsingSplitAndClassifiers("SPLIT POWSPECTRUM", "Classifier trainer", selectedFeats, epochCount[0], destFile)
                    modifyTrainScenUsingSplitAndClassifiers("SPLIT CONNECT", "Classifier trainer", selectedFeats2, epochCount[1], destFile)
                    modifyConnectivityMetric(self.extractDict["ConnectivityMetric"], destFile)

            elif i == 4 and not self.usingDualFeatures and self.parameterDict["pipelineType"] == optionKeys[2]:
                # "speed up" training scenarios (ONLY CONNECTIVITY)
                modifyTrainScenUsingSplitAndCsvWriter("SPLIT", selectedFeats, epochCount[0], destFile, trainingpath)

            elif i == 3:
                #  "online" scenario
                modifyAcqScenario(destFile, self.parameterDict["AcquisitionParams"], True)
                if not self.usingDualFeatures:
                    modifyTrainScenUsingSplitAndClassifiers("SPLIT", "Classifier processor", selectedFeats, epochCount[0], destFile)
                    # Special case: "connectivity metric"
                    if "ConnectivityMetric" in self.extractDict.keys():
                        modifyConnectivityMetric(self.extractDict["ConnectivityMetric"], destFile)
                else:
                    modifyTrainScenUsingSplitAndClassifiers("SPLIT POWSPECTRUM", "Classifier processor", selectedFeats, epochCount[0], destFile)
                    modifyTrainScenUsingSplitAndClassifiers("SPLIT CONNECT", "Classifier processor", selectedFeats2, epochCount[1], destFile)
                    modifyConnectivityMetric(self.extractDict["ConnectivityMetric"], destFile)



        # ------------------------------------------
        # TEST USING NEW SPLITTED "SPED UP" TRAINING
        # In this experimental setup, we want to avoid re-computing connectivity matrices
        # each time a training is attempted.
        # To this end, we use the connectivity matrices (as node strengths) computed in
        # the extraction step. For each run selected for the training, we compute a "temporary"
        # feature vector, and store it before we would feed it to the "classifier trainer" box in
        # openvibe.
        # When all runs have been processed that way, we aggregate all feature vectors (per class, per
        # feat), and feed those composite files to the feature aggregators and classifier trainer (in
        # scenario sc2-train-speedup-finalize.xml)
        if self.speedUp and self.parameterDict["pipelineType"] == optionKeys[2]:

            # "First step"
            scenFile = os.path.join(self.workspaceFolder, templateScenFilenames[4])
            analysisPath = os.path.join(self.workspaceFolder, "sessions", self.currentSessionId, "extract")
            trainingPath = os.path.join(self.workspaceFolder, "sessions", self.currentSessionId, "train")

            self.info2.emit("Running first step (feature vector computation per run)")
            for run in trainingSigList:
                # MODIFY "FIRST STEP" SCENARIO
                runBasename = run.split("\\")[-1].removesuffix("-TRIALS.csv")
                modifyTrainingFirstStep(runBasename, len(selectedFeats), analysisPath, trainingPath, self.currentSessionId, scenFile)

                # RUN THE SCENARIO, with "False" to ignore training results
                success = self.runOvScenarioGeneric(os.path.join(self.workspaceFolder, scenFile))
                if not success:
                    successGlobal = False
                    self.errorMessageTrainer()
                    self.over.emit(False, self.exitText)
                    break

            # NOW AGGREGATE FEATURE VECTORS PER FEATURE AND PER CLASS...
            self.info.emit(True)
            self.info2.emit("Combining Feature Vectors...")
            compositeFiles = []
            for classIdx in [1, 2]:
                for feat in range(len(selectedFeats)):
                    featFilesToMerge = []
                    for run in trainingSigList:
                        runBasename = run.split("\\")[-1].removesuffix("-TRIALS.csv")
                        runFeatClassCmb = str(runBasename + "-class" + str(classIdx) + "-feat" + str(feat+1) + ".csv" )
                        if os.path.exists(os.path.join(trainingPath, self.currentSessionId, runFeatClassCmb)):
                            featFilesToMerge.append(os.path.join(trainingPath, self.currentSessionId, runFeatClassCmb))

                    outCsv = mergeRunsCsv_new(featFilesToMerge)
                    compositeFiles.append(outCsv)

            # MODIFY "SECOND STEP" SCENARIO INPUTS & OUTPUT...
            scenFile = os.path.join(self.workspaceFolder, templateScenFilenames[5])
            newWeightsName = str("classifier-weights-" + self.attemptId + ".xml")
            modifyTrainingSecondStep(compositeFiles, len(selectedFeats), newWeightsName, self.currentSessionId, scenFile)
            modifyTrainPartitions(self.trainingSize, scenFile)

            self.info2.emit("Finalizing Training...")
            scenXml = os.path.join(self.workspaceFolder, templateScenFilenames[5])
            success, classifierOutputStr, accuracy = self.runClassifierScenario(scenXml)
            if not success:
                successGlobal = False
                self.errorMessageTrainer()
                self.over.emit(False, self.exitText)
            else:
                # Copy weights file to <workspaceFolder>/classifier-weights.xml
                newWeights = os.path.join(self.signalFolder, "sessions", self.currentSessionId, \
                                          "train", str("classifier-weights-" + self.attemptId + ".xml"))
                origFilename = os.path.join(self.workspaceFolder, "classifier-weights.xml")
                copyfile(newWeights, origFilename)

                self.currentAttempt["Score"] = accuracy

                # PREPARE GOODBYE MESSAGE...
                textFeats = str("")
                for i in range(len(trainingSigList)):
                    textFeats += str(os.path.basename(trainingSigList[i]) + "\n")

                textFeats += str("\nFeature(s) ")
                textFeats += str("(" + self.parameterDict["pipelineType"] + "):\n")
                for i in range(len(selectedFeats)):
                    textFeats += str(
                        "\t" + "Channel " + str(selectedFeats[i][0]) + " at " + str(selectedFeats[i][1]) + " Hz\n")

                textDisplay = textFeats
                textDisplay += str("\n" + classifierOutputStr)

                self.exitText = textDisplay

        else:
            # ORIGINAL VERSION - NOT SPED UP
            # USING "CLASSIC" COMPOSITE FILE BUILDING + RUNNING TRAIN SCEN ON IT
            scenFile = os.path.join(self.workspaceFolder, templateScenFilenames[2])
            modifyTrainPartitions(self.trainingSize, scenFile)

            # Run the first training scenario  composite file from selected items
            class1Stim = "OVTK_GDF_Left"
            class2Stim = "OVTK_GDF_Right"
            tmin = 0
            tmax = float(self.extractDict["StimulationEpoch"])

            # Train on accumulation of all selected trials
            compositeCsv = mergeRunsCsv(trainingSigList, self.parameterDict["AcquisitionParams"]["Class1"],
                                        self.parameterDict["AcquisitionParams"]["Class2"],
                                        class1Stim, class2Stim, tmin, tmax)
            if not compositeCsv:
                self.over.emit(False, "Error merging runs!! Most probably different list of electrodes")
                return

            print("Composite file for training: " + compositeCsv)
            compositeCsvBasename = os.path.basename(compositeCsv)
            newWeightsName = str("classifier-weights-" + self.attemptId + ".xml")
            modifyTrainIO(compositeCsvBasename, newWeightsName, self.currentSessionId, scenFile)

            # write composite file name in structure for future saving in workspace
            self.currentAttempt["CompositeFile"] = compositeCsvBasename

            # increment progressbar
            self.info.emit(True)
            self.info2.emit("Running Training Scenario")

            # RUN THE CLASSIFIER TRAINING SCENARIO
            scenXml = os.path.join(self.workspaceFolder, templateScenFilenames[2])
            success, classifierOutputStr, accuracy = self.runClassifierScenario(scenXml)

            if not success:
                self.errorMessageTrainer()
                self.over.emit(False, self.exitText)
            else:
                # Copy weights file to <workspaceFolder>/classifier-weights.xml
                newWeights = os.path.join(self.workspaceFolder, "sessions", self.currentSessionId, "train", \
                                          str("classifier-weights-" + self.attemptId + ".xml"))
                origFilename = os.path.join(self.workspaceFolder, "classifier-weights.xml")
                copyfile(newWeights, origFilename)

                # write score in workspace structure for future reporting
                self.currentAttempt["Score"] = accuracy

                # PREPARE GOODBYE MESSAGE...
                textFeats = str("")
                for i in range(len(trainingSigList)):
                    textFeats += str(os.path.basename(trainingSigList[i]) + "\n")

                if not self.usingDualFeatures:
                    textFeats += str("\nFeature(s) ")
                    textFeats += str("(" + self.parameterDict["pipelineType"]+"):\n")
                    for i in range(len(selectedFeats)):
                        textFeats += str("\t"+"Channel " + str(selectedFeats[i][0]) + " at " + str(selectedFeats[i][1]) + " Hz\n")
                else:
                    textFeats += str("\nFeature(s) for PowSpectrum:\n")
                    for i in range(len(selectedFeats)):
                        textFeats += str(
                            "\t" + "Channel " + str(selectedFeats[i][0]) + " at " + str(selectedFeats[i][1]) + " Hz\n")
                    textFeats += str("Feature(s) for Connectivity:\n")
                    for i in range(len(selectedFeats2)):
                        textFeats += str(
                            "\t" + "Channel " + str(selectedFeats2[i][0]) + " at " + str(selectedFeats2[i][1]) + " Hz\n")

                textDisplay = textFeats
                textDisplay += str("\n" + classifierOutputStr)

                self.exitText = textDisplay


        self.stop = True
        self.over.emit(True, self.exitText)

    def stopThread(self):
        self.stop = True

    def errorMessageTrainer(self):
        textError = str("Error running \"Training\" scenario\n")
        textError += str("Please try again with a lower number of partitions for k-fold test\n")
        self.exitText = textError

    def checkSelectedFeats(self, inputSelectedFeats, sampFreq, electrodeList):
        selectedFeats = []
        errMsg = ""
        # Checks :
        # - No empty field
        # - frequencies in acceptable ranges
        # - channels in list
        n_bins = int((sampFreq / 2) + 1)
        for idx, feat in enumerate(inputSelectedFeats):
            if feat == "":
                errMsg = str("Pair " + str(idx + 1) + " is empty...")
                return None, errMsg
            [chan, freqstr] = feat.split(";")
            if chan not in electrodeList:
                errMsg = str("Channel in pair " + str(idx + 1) + " (" + str(chan) + ") is not in the list...")
                return None, errMsg
            freqs = freqstr.split(":")
            for freq in freqs:
                if not freq.isdigit():
                    errMsg = str("Frequency in pair " + str(idx + 1) + " (" + str(
                        freq) + ") has an invalid format, must be an integer...")
                    return None, errMsg
                if int(freq) >= n_bins:
                    errMsg = str(
                        "Frequency in pair " + str(idx + 1) + " (" + str(freq) + ") is not in the acceptable range...")
                    return None, errMsg
            selectedFeats.append(feat.split(";"))
            print(feat)

        return selectedFeats, errMsg

    def runClassifierScenario(self, scenFile):
        # ----------
        # Run the provided training scenario, and check the console output for termination, errors, and
        # classification results
        # ----------

        # TODO WARNING : MAYBE CHANGE THAT IN THE FUTURE...
        # enableConfChange = False
        # if enableConfChange:
        #     # CHECK IF openvibe.conf has randomization of k-fold enabled
        #     # if not, change it
        #     confFile = os.path.join(os.path.dirname(self.ovScript), "share", "openvibe", "kernel", "openvibe.conf")
        #     if platform.system() == 'Windows':
        #         confFile = confFile.replace("/", "\\")
        #     modifyConf = False
        #     with open(confFile, 'r') as conf:
        #         confdata = conf.read()
        #         if "Plugin_Classification_RandomizeKFoldTestData = false" in confdata:
        #             modifyConf = True
        #             confdata = confdata.replace("Plugin_Classification_RandomizeKFoldTestData = false", "Plugin_Classification_RandomizeKFoldTestData = true")
        #     if modifyConf:
        #         with open(confFile, 'w') as conf:
        #             conf.write(confdata)

        # BUILD THE COMMAND (use designer.cmd from GUI)
        command = self.ovScript
        if platform.system() == 'Windows':
            command = command.replace("/", "\\")

        # Run actual command (openvibe-designer.cmd --no-gui --play-fast <scen.xml>)
        p = subprocess.Popen([command, "--invisible", "--play-fast", scenFile],
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        # Read console output to detect end of process
        # and prompt user with classification score. Quite artisanal but works
        success = True
        classifierOutputStr = ""
        activateScoreMsgBox = False
        while True:
            output = p.stdout.readline()
            if p.poll() is not None:
                break
            if output:
                print(str(output))
                if "Invalid indexes: stopIdx - trainIndex = 1" in str(output):
                    success = False
                    return success, None, None
                if "Application terminated" in str(output):
                    classifierOutputStr = str(classifierOutputStr + "\n")
                    break
                if "Cross-validation test" in str(output):
                    activateScoreMsgBox = True
                if activateScoreMsgBox:
                    stringToWrite = str(output).replace("\\r\\n\'", "")
                    if "trainer>" in stringToWrite:
                        stringToWrite = stringToWrite.split("trainer> ")
                        classifierOutputStr = str(classifierOutputStr + stringToWrite[1] + "\n")

        if activateScoreMsgBox:
            lines = classifierOutputStr.splitlines()

            target_1_True_Negative = float(lines[2].split()[2])
            target_1_False_Positive = float(lines[2].split()[3])
            target_2_False_Negative = float(lines[3].split()[2])
            target_2_True_Positive = float(lines[3].split()[3])

            precision_Class_1 = round(target_1_True_Negative / (target_1_True_Negative + target_2_False_Negative), 2)
            sensitivity_Class_1 = round(target_1_True_Negative / (target_1_True_Negative + target_1_False_Positive), 2)
            precision_Class_2 = round(target_2_True_Positive / (target_2_True_Positive + target_1_False_Positive), 2)
            sensitivity_Class_2 = round(target_2_True_Positive / (target_2_True_Positive + target_2_False_Negative), 2)

            accuracy = round(100.0 * (target_1_True_Negative + target_2_True_Positive) / (
                    target_1_True_Negative + target_1_False_Positive + target_2_False_Negative + target_2_True_Positive),
                             2)

            F_1_Score_Class_1 = round(
                2 * precision_Class_1 * sensitivity_Class_1 / (precision_Class_1 + sensitivity_Class_1), 2)
            F_1_Score_Class_2 = round(
                2 * precision_Class_2 * sensitivity_Class_2 / (precision_Class_2 + sensitivity_Class_2), 2)

            messageClassif = "Overall accuracy : " + str(accuracy) + "%\n"
            messageClassif += "Class 1 | Precision  : " + str(precision_Class_1) + " | " + "Sensitivity : " + str(
                sensitivity_Class_1)
            messageClassif += " | F_1 Score : " + str(F_1_Score_Class_1) + "\n"
            messageClassif += "Class 2 | Precision  : " + str(precision_Class_2) + " | " + "Sensitivity : " + str(
                sensitivity_Class_2)
            messageClassif += " | F_1 Score : " + str(F_1_Score_Class_2)

            return success, messageClassif, accuracy

        else:
            return success

    def runOvScenarioGeneric(self, scenXml):
        # ----------
        # Run a specified OpenViBE scenario. Filename must be absolute.
        # The console output is read only to find "Application terminated" or specific errors
        # ----------

        # BUILD THE COMMAND (use designer.cmd from GUI)
        command = self.ovScript
        if platform.system() == 'Windows':
            command = command.replace("/", "\\")

        # Run actual command (openvibe-designer.cmd --no-gui --play-fast <scen.xml>)
        p = subprocess.Popen([command, "--invisible", "--play-fast", scenXml],
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        # Read console output to detect end of process
        # and prompt user with classification score. Quite artisanal but works
        success = True
        outputStr = ""
        activateScoreMsgBox = False
        while True:
            output = p.stdout.readline()
            if p.poll() is not None:
                break
            if output:
                print(str(output))
                if "Invalid indexes: stopIdx - trainIndex = 1" in str(output):
                    success = False
                    return success
                if "Application terminated" in str(output):
                    outputStr = str(outputStr + "\n")
                    break

        return success




from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QProgressBar
from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout
from PyQt5.QtCore import Qt, QBasicTimer, QThread, pyqtSignal

import sys

class ProgressBar(QDialog):

    def __init__(self, title, label, nbIncrements, parent=None):

        super().__init__(parent)

        self.nbIncrements = nbIncrements
        self.resize(800, 150)
        self.setWindowTitle(title)
        self.currentProgress = 0

        self.label = QLabel(self)
        self.label.setText(label)
        self.label.setAlignment(Qt.AlignCenter)

        self.featProgressBar = QProgressBar(self)
        self.featProgressBar.setTextVisible(True)
        self.featProgressBar.setMinimum(0)
        self.featProgressBar.setMaximum(nbIncrements)
        self.featProgressBar.setValue(0)
        self.featProgressBar.setFormat(str(self.currentProgress) + "/" + str(self.nbIncrements))

        layout = QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.featProgressBar)
        self.setLayout(layout)
        self.show()

    def increment(self):
        self.currentProgress += 1
        self.featProgressBar.setValue(self.currentProgress)
        self.featProgressBar.setFormat(str(self.currentProgress) + "/" + str(self.nbIncrements))

    def changeLabel(self, newLabel):
        self.label.setText(newLabel)

    def finish(self):
        self.close()

    def showBar(self):
        self.show()

class ProgressBarNoInfo(QDialog):

    def __init__(self, label, parent=None):

        super().__init__(parent)

        self.resize(350, 100)
        self.setWindowTitle("Processing...")

        self.tipLabel = QLabel(label)

        self.featProgressBar = QProgressBar(self)
        self.featProgressBar.setRange(0, 0)

        tipLayout = QHBoxLayout()
        tipLayout.addWidget(self.tipLabel)

        featLayout = QHBoxLayout()
        featLayout.addWidget(self.featProgressBar)

        layout = QVBoxLayout()
        layout.addLayout(featLayout)
        layout.addLayout(tipLayout)
        self.setLayout(layout)
        self.show()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    label = "Extracting features"
    filenum = 10
    progress = ProgressBar(label, filenum)
    progress.show()
    app.exec_()



import numpy as np
from scipy import signal,stats,fft
from spectrum import arburg,arma2psd,pburg
import statsmodels.regression.linear_model as transform


def Power_burg_calculation_optimization(Epoch_compute,noverlap,N_FFT,f_max, n_per_seg,smoothing,freqs_left,filter_order):
    #burg = pburg(Epoch_compute,15,NFFT = nfft_test)
    data = Epoch_compute[:,:]
    xs, ys = [], []
    a = data.shape
    M = a[1]
    #print(M)
    L = n_per_seg
    #print(L)
    noverlap = noverlap
    LminusOverlap = L-noverlap
    #print(LminusOverlap)
    k = round((M-noverlap)/(L-noverlap))
    #print(k)
    xStart = np.array(range(0,k*LminusOverlap,LminusOverlap))
    #print(xStart)
    xEnd = []
    for i in xStart:
        xEnd.append(i+L-1)
    xEnd = np.array(xEnd)
    #print(xEnd)
    fres =f_max/N_FFT
    power = []
    Block_spectrum= []
    
    tab = np.array(range(k-1))
    trialspectrum = np.zeros([a[0],N_FFT])
    PSD_final = np.zeros([a[0],round(f_max/2)])
    Time_freq = np.zeros([a[0],len(tab),round(f_max/(2*fres))])
    time = np.linspace(0,round(M/f_max),k-1)
    for i in range(a[0]):
        Block_spectrum = []
        for numBlock in tab:
            aux_condition1data = Epoch_compute[i,xStart[numBlock]:xEnd[numBlock]]

            aux_condition1data = signal.detrend(aux_condition1data,type='constant')

            AR, sigma2 = transform.burg(aux_condition1data, filter_order)
            PSD = arma2psd(-AR, NFFT = N_FFT,sides='centerdc')
            # plt.plot(PSD)
            # plt.show()
            Block_spectrum.append(PSD)
            Time_freq[i,numBlock]=PSD[round(f_max/(2*fres)):round(f_max/fres)]
            block = np.mean(Block_spectrum,axis=0)
            trialspectrum[i] = np.array(block)

    if smoothing == True:
        for k in range(a[0]):
             for l in range(a[1]):
                 PSD_final[k,l,0] = (trialspectrum[k,l,0] + trialspectrum[k,l,1] +trialspectrum[k,l,2])/3
                 PSD_final[k,l,round(f_max/2)-1] = (trialspectrum[k,l,freqs_left.shape[0]-3] + trialspectrum[k,l,freqs_left.shape[0]-2] +trialspectrum[k,l,freqs_left.shape[0]-1])/3
                 for i in range(5,freqs_left.shape[0]-2,round(1/fres)):
                     PSD_final[k,l,round(i/5)] = (trialspectrum[k,l,i-2] +trialspectrum[k,l,i-1] +trialspectrum[k,l,i] + trialspectrum[k,l,i+1] +trialspectrum[k,l,i+2])/(5)
        return PSD_final

    return trialspectrum[:,round(f_max/(2*fres)):round(f_max/fres)],Time_freq,time


def Power_burg_calculation(Epoch_compute,noverlap,N_FFT,f_max, n_per_seg,smoothing,freqs_left,filter_order):
    #burg = pburg(Epoch_compute,15,NFFT = nfft_test)
    data = Epoch_compute[:,:,:]
    xs, ys = [], []
    a = data.shape
    M = a[2]
    #print(M)
    L = n_per_seg
    #print(L)
    noverlap = noverlap
    LminusOverlap = L-noverlap
    #print(LminusOverlap)
    k = round((M-noverlap)/(LminusOverlap))
    #print(k)
    xStart = np.array(range(0,k*LminusOverlap,LminusOverlap))
    #print(xStart)
    xEnd = []
    for i in xStart:
        xEnd.append(i+L-1)
    xEnd = np.array(xEnd)
    #print(xEnd)
    
    fres =f_max/N_FFT
    power = []
    Block_spectrum= []
    
    tab = np.array(range(k-1))
    trialspectrum = np.zeros([a[0],a[1],N_FFT])
    PSD_final = np.zeros([a[0],a[1],round(f_max/2)])
    Time_freq = np.zeros([a[0],a[1],len(tab),round(f_max/(2*fres))])
    time = np.linspace(0,round(M/f_max),k-1)
    for i in range(a[0]):
        for j in range(a[1]):
            Block_spectrum = []
            for numBlock in tab:
                aux_condition1data = Epoch_compute[i,j,xStart[numBlock]:xEnd[numBlock]]

                aux_condition1data = signal.detrend(aux_condition1data,type='constant')

                AR, sigma2 = transform.burg(aux_condition1data, filter_order)
                PSD = arma2psd(-AR, NFFT = N_FFT,sides='centerdc')
                # plt.plot(PSD)
                # plt.show()
                Block_spectrum.append(PSD)

                Time_freq[i,j,numBlock]=PSD[round(f_max/(2*fres)):round(f_max/fres)]
            block = np.mean(Block_spectrum,axis=0)
            trialspectrum[i,j] = np.array(block)

    if smoothing == True:
        for k in range(a[0]):
             for l in range(a[1]):
                 PSD_final[k,l,0] = (trialspectrum[k,l,0] + trialspectrum[k,l,1] +trialspectrum[k,l,2])/3
                 PSD_final[k,l,round(f_max/2)-1] = (trialspectrum[k,l,freqs_left.shape[0]-3] + trialspectrum[k,l,freqs_left.shape[0]-2] +trialspectrum[k,l,freqs_left.shape[0]-1])/3
                 for i in range(5,freqs_left.shape[0]-2,round(1/fres)):
                     PSD_final[k,l,round(i/5)] = (trialspectrum[k,l,i-2] +trialspectrum[k,l,i-1] +trialspectrum[k,l,i] + trialspectrum[k,l,i+1] +trialspectrum[k,l,i+2])/(5)
        return PSD_final

    return trialspectrum[:,:,round(f_max/(2*fres)):round(f_max/fres)],Time_freq,time

def Power_calculation_welch_method(Epoch_compute,f_min,f_max,t_min,t_max,nfft,noverlap,nper_seg,pick,proje,averag,windowing, smoothing):
    #filtered = mne.filter.filter_data(Epoch_compute, 140, 7, 35)

    fres =f_max/nfft
    #psd_left,freqs_left = mne.time_frequency.psd_welch(Epoch_compute, fmin=f_min, fmax=f_max, tmin=t_min, tmax=t_max, n_fft=nfft, n_overlap=noverlap, n_per_seg=nper_seg, picks=pick, proj=proje, n_jobs=1, reject_by_annotation=True, average=averag, window=windowing, verbose=None)
    freqs_left,psd_left = signal.welch(Epoch_compute, fs=f_max, window=windowing, nperseg=nper_seg, noverlap=noverlap, nfft=nfft, detrend='constant', return_onesided=True, scaling='density', axis=- 1, average=averag)
    b = psd_left.shape
    #print(freqs_left.shape[0])
    PSD_final = np.zeros([b[0],b[1],round(f_max/2)])
    if smoothing == True:
        for k in range(b[0]):
             for l in range(b[1]):
                 PSD_final[k,l,0] = (psd_left[k,l,0] + psd_left[k,l,1] +psd_left[k,l,2])/3
                 PSD_final[k,l,round(f_max/2)-1] = (psd_left[k,l,freqs_left.shape[0]-3] + psd_left[k,l,freqs_left.shape[0]-2] +psd_left[k,l,freqs_left.shape[0]-1])/3
                 for i in range(5,freqs_left.shape[0]-2,round(1/fres)):
                     PSD_final[k,l,round(i/5)] = (psd_left[k,l,i-2] +psd_left[k,l,i-1] +psd_left[k,l,i] + psd_left[k,l,i+1] +psd_left[k,l,i+2])/(5)
        return PSD_final,freqs_left

                #avgdata2(i, :, countcond2) = mean(trialspectrum(ind-evaluationsPerBin/2:ind+evaluationsPerBin/2-1,:));
    #print(PSD_final.shape)
    return psd_left,freqs_left



import sys
import os
import mne
import subprocess
from multiprocessing import Process
from shutil import copyfile
import json
from threading import Thread
import pandas as pd
from importlib import resources

from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QFormLayout
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QLineEdit

from PyQt5.QtGui import QFont

from happyfeat.lib.modifyOpenvibeScen import *
import happyfeat.lib.bcipipeline_settings as settings
from happyfeat.lib.utils import *
import happyfeat.lib.workspaceMgmt as workspaceMgmt

import happyfeat.featureExtractionInterface as featExtractApp

class Dialog(QDialog):

    def __init__(self, workspace, parent=None):

        super().__init__(parent)

        # GENERAL SETTINGS (WORKSPACES, TEMPLATES...)
        self.launchTrue = False
        self.ovScript = None    # to be r/w in config.json
        self.customMontagePath = None
        self.templateFolder = None
        self.workspace = workspace  # fullpath of the .hfw file
        self.workspaceFolder = os.path.splitext(self.workspace)[0]  # only the folder (still fullpath)
        if not os.path.exists(self.workspace) or not os.path.exists(self.workspaceFolder):
            myMsgBox("Specified workspace does not exist. Please use the \"Welcome\" GUI first!")
            self.reject()

        # Get user's config
        self.userConfig = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.json')
        if not os.path.exists(self.userConfig):
            myMsgBox("General config file config.json not found! Please use happyfeat_welcome.py script.")
            self.reject()
        with open(self.userConfig, "r+") as userCfg:
            currentCfg = json.load(userCfg)
            if "lastWorkspacePath" in currentCfg:
                self.workspacesFolder = currentCfg["lastWorkspacePath"]
            if "ovScript" in currentCfg:
                self.ovScript = currentCfg["ovScript"]

        # SCENARIO PARAMETERS...
        self.parameterDict = {}
        self.parameterTextList = []  # for parsing later...

        # INTERFACE INIT...
        self.setWindowTitle('HappyFeat - Acquisition parameters GUI')
        self.dlgLayout = QVBoxLayout()

        # SPECIAL LAYOUTS...
        self.montageLayoutIdx = None
        self.customMontageLayoutIdx = None

        # Start listing the widgets, we'll add them to the layout at the end
        label = str("=== Protocol Selection ===")
        self.label = QLabel(label)
        self.label.setFont(QFont("system-ui", 12))
        self.label.setAlignment(QtCore.Qt.AlignCenter)

        self.scenarioComboBox = QComboBox(self)
        for idx, key in enumerate(settings.optionKeys):
            self.scenarioComboBox.addItem(settings.optionsComboText[key], idx)
        # self.scenarioComboBox.currentIndexChanged.connect(self.scenarioComboBoxChanged)

        # COLUMN OF PARAMETERS
        self.vBoxLayout = QVBoxLayout()
        self.vBoxLayout.setAlignment(QtCore.Qt.AlignTop)

        # ACQ PARAMETERS
        self.paramWidgets = []
        self.initAcqParams()

        # OpenViBE designer executable...
        labelOv = QLabel()
        labelOvtxt = str("OpenViBE designer script (openvibe-designer.cmd or .exe or .sh)")
        labelOvtxt = str(labelOvtxt + "\n(in your OpenViBE installation folder)")
        labelOv.setText(labelOvtxt)
        labelOv.setAlignment(QtCore.Qt.AlignCenter)
        self.vBoxLayout.addWidget(labelOv)

        self.btn_browseOV = QPushButton("Browse...")
        self.btn_browseOV.clicked.connect(lambda: self.browseForDesigner())
        self.designerWidget = QWidget()
        layout_h = QHBoxLayout(self.designerWidget)
        self.designerTextBox = QLineEdit()
        self.designerTextBox.setText(self.ovScript)
        self.designerTextBox.setEnabled(False)
        layout_h.addWidget(self.designerTextBox)
        layout_h.addWidget(self.btn_browseOV)
        self.vBoxLayout.addWidget(self.designerWidget)

        # Generate button
        self.btn_generateLaunch = QPushButton("Generate scenarios, launch HappyFeat")
        self.btn_generateLaunch.clicked.connect(lambda: self.generate(True))
        # self.btn_generate = QPushButton("Generate scenarios, let me handle things!")
        # self.btn_generate.clicked.connect(lambda: self.generate(False))

        # Add all widgets to the general layout...
        self.dlgLayout.addWidget(self.label)
        self.dlgLayout.addWidget(self.scenarioComboBox)
        self.dlgLayout.addLayout(self.vBoxLayout)
        self.dlgLayout.addWidget(self.btn_generateLaunch)
        # self.dlgLayout.addWidget(self.btn_generate)

        # Display layout
        self.setLayout(self.dlgLayout)
        self.show()

    def initAcqParams(self):

        # PARAMETERS
        labelText = [None, None]
        label = [None, None]
        labelText[0] = str("=== ")
        labelText[0] = str(labelText[0] + "Acquisition and Stimulation")
        labelText[0] = str(labelText[0] + " ===")
        label[0] = QLabel(labelText[0])
        label[0].setAlignment(QtCore.Qt.AlignCenter)
        label[0].setFont(QFont("system-ui", 12))
        self.vBoxLayout.addWidget(label[0])

        # PARAMETER ALWAYS PRESENT : ELECTRODES MONTAGE
        # Drop-down menu with montages available in MNE
        # + Custom option (will open a file selection in the future... TODO)
        self.layoutMontages = QVBoxLayout()
        self.layoutMontageSelectionH = QHBoxLayout()
        self.layoutMontageV1 = QVBoxLayout()
        self.layoutMontageV2 = QVBoxLayout()
        labelMontage = QLabel()
        labelMontage.setText("Channel Montage")
        self.layoutMontageV1.addWidget(labelMontage)
        self.layoutMontageV1.setAlignment(QtCore.Qt.AlignTop)
        self.montageComboBox = QComboBox(self)
        montageIdx = 0
        self.montageComboBox.addItem("Custom... (select file)")
        for idx, mtg in enumerate(mne.channels.get_builtin_montages()):
            self.montageComboBox.addItem(mtg)
            if mtg == "biosemi64":  # default
                montageIdx = idx + 1

        self.montageComboBox.setCurrentIndex(montageIdx)
        self.montageComboBox.currentIndexChanged.connect(self.montageComboBoxChanged)
        self.layoutMontageV2.addWidget(self.montageComboBox)
        self.layoutMontageSelectionH.addLayout(self.layoutMontageV1)
        self.layoutMontageSelectionH.addLayout(self.layoutMontageV2)
        self.layoutMontages.addLayout(self.layoutMontageSelectionH)

        self.vBoxLayout.addLayout(self.layoutMontages)

        # OTHER PARAMETERS

        formLayout = None
        formLayout = QFormLayout()

        # GET PARAMETER LIST FOR SELECTED BCI PIPELINE, AND DISPLAY THEM
        for idx, param in enumerate(settings.pipelineAcqSettings):
            # init params...
            value = settings.pipelineAcqSettings[param]
            self.parameterDict[param] = value
            # create widgets...
            paramWidget = QLineEdit()
            paramWidget.setText(str(value))
            settingLabel = settings.paramIdText[param]
            self.paramWidgets.append(paramWidget)
            self.parameterTextList.append(param)
            formLayout.addRow(settingLabel, self.paramWidgets[-1])

        self.vBoxLayout.addLayout(formLayout)

        return

    def montageComboBoxChanged(self, ix):
        if ix == 0:  # Custom
            # open new layout to let user choose montage file
            self.btn_browseMontage = QPushButton("Browse...")
            self.btn_browseMontage.clicked.connect(lambda: self.browseForMontage())
            self.customMontageWidget = QWidget()
            layout_h = QHBoxLayout(self.customMontageWidget)
            self.customMontageTxtBox = QLineEdit()
            self.customMontageTxtBox.setText("")
            self.customMontageTxtBox.setEnabled(False)
            layout_h.addWidget(self.customMontageTxtBox)
            layout_h.addWidget(self.btn_browseMontage)
            self.layoutMontageV2.addWidget(self.customMontageWidget)

        else:
            # if "custom" layout had been opened before, remove it.
            if self.layoutMontageV2.count() > 1:
                self.btn_browseMontage.setParent(None)
                self.customMontageWidget.setParent(None)
                self.customMontageTxtBox.setParent(None)
                self.layoutMontageV2.removeItem(self.layoutMontageV2.itemAt(1))

        return

    def browseForDesigner(self):
        directory = os.getcwd()
        self.ovScript, dummy = QFileDialog.getOpenFileName(self, "OpenViBE designer script", str(directory))
        if "openvibe-designer.cmd" or "openvibe-designer.exe" or "openvibe-designer.sh" in self.ovScript:
            self.designerTextBox.setText(self.ovScript)
            workspaceMgmt.setKeyValue(self.userConfig, "ovScript", self.ovScript)
        else:
            self.ovScript = None

        return

    def browseForMontage(self):
        directory = os.getcwd()
        self.customMontagePath, dummy = QFileDialog.getOpenFileName(self, "Open electrode names file", str(directory))
        if self.checkCustomMontage():
            self.customMontageTxtBox.setText(self.customMontagePath)

        return

    def checkCustomMontage(self):
        invalid = False
        column_names = ['name', 'x', 'y', 'z']
        csv = pd.read_csv(self.customMontagePath)
        if len(csv.columns) != len(column_names):
            myMsgBox("Invalid channel locations file. Please refer to the documentation.")
            return False
        if not all(csv.columns[x] == column_names[x] for x in range(len(column_names))):
            myMsgBox("Invalid channel locations file. Please refer to the documentation.")
            return False

        return True

    def generate(self, launch):

        if self.ovScript == "" or not self.ovScript:
            myMsgBox("Please enter a correct path for the OpenViBE designer")
            return

        ix = self.scenarioComboBox.currentIndex()
        if ix == 0:
            myMsgBox("Please select a protocol...")
            return

        pipelineKey = settings.optionKeys[ix]

        self.templateFolder = settings.optionsTemplatesDir[pipelineKey]
        print(str("TEMPLATE FOLDER : " + self.templateFolder))

        self.parameterDict["pipelineType"] = pipelineKey

        ####
        # FIRST STEP : CREATE PARAMETER DICTIONARY
        ###

        # OpenViBE Path...
        if self.ovScript is None:
            myMsgBox("Please enter a valid path for the openViBE designer script")
            return

        if self.montageComboBox.currentIndex() == 0:  # custom
            if self.customMontagePath is None:
                myMsgBox("Please provide a valid file for the custom montage")
                return

        # Montage : default values...
        self.parameterDict["sensorMontage"] = self.montageComboBox.currentText()
        self.parameterDict["customMontagePath"] = ''
        if self.montageComboBox.currentIndex() == 0:
            self.parameterDict["sensorMontage"] = "custom"
            self.parameterDict["customMontagePath"] = self.customMontagePath

        # self.parameterDict["ChannelNames"] = electrodes
        self.parameterDict["ovDesignerPath"] = self.ovScript

        self.parameterDict["Sessions"] = {}

        # Acquisition parameters, set in this GUI...
        self.parameterDict["AcquisitionParams"] = {}
        for i in range(len(self.paramWidgets)):
            param = self.parameterTextList[i]
            if param in self.parameterDict:
                # TODO : remove duplication... (done to make params.json and workspace file coexist)
                self.parameterDict["AcquisitionParams"][param] = self.paramWidgets[i].text()
                self.parameterDict[param] = self.paramWidgets[i].text()

        # Write default parameters for selected pipeline
        extractParamDict = settings.pipelineExtractSettings[self.parameterDict["pipelineType"]].copy()

        self.parameterDict["currentSessionId"] = "1"
        print(self.parameterDict)

        # WRITE WORKSPACE FILE
        if not self.workspace:
            # TODO !! manage error
            myMsgBox("Missing workspace file!!")
        else:
            workspaceMgmt.setKeyValue(self.workspace, "ovDesignerPath", self.parameterDict["ovDesignerPath"])
            workspaceMgmt.setKeyValue(self.workspace, "pipelineType", self.parameterDict["pipelineType"])
            workspaceMgmt.setKeyValue(self.workspace, "sensorMontage", self.parameterDict["sensorMontage"])
            workspaceMgmt.setKeyValue(self.workspace, "customMontagePath", self.parameterDict["customMontagePath"])
            workspaceMgmt.setKeyValue(self.workspace, "AcquisitionParams", self.parameterDict["AcquisitionParams"])
            workspaceMgmt.setKeyValue(self.workspace, "currentSessionId", self.parameterDict["currentSessionId"])
            workspaceMgmt.newSession(self.workspace, self.parameterDict, "1", extractParamDict)

        # GENERATE (list of files in settings.templateScenFilenames)
        #   SC1 (ACQ/MONITOR)
        #   SC2 (FEATURE EXT)
        #   SC2 (TRAIN)
        #   SC3 (ONLINE)
        #   SC2-SPEEDUP-FIRSTSTEP (TRAIN+, 1)
        #   SC2-SPEEDUP-FINALIZE  (TRAIN+, 2)
        for filename in settings.templateScenFilenames:
            with resources.path(str(__name__.split('.')[0] + '.' + self.templateFolder), filename) as srcFile:
                destFile = os.path.join(self.workspaceFolder, filename)
                if os.path.exists(srcFile):
                    print("---Copying file " + str(srcFile) + " to " + str(destFile))
                    copyfile(srcFile, destFile)
                    if "xml" in destFile:
                        modifyScenarioGeneralSettings(destFile, self.parameterDict)

        # SPECIAL CASES :
        #   SC1 & SC3 : "GRAZ" BOX SETTINGS
        modifyAcqScenario(os.path.join(self.workspaceFolder,
                                       settings.templateScenFilenames[0]), self.parameterDict, False)
        modifyAcqScenario(os.path.join(self.workspaceFolder,
                                       settings.templateScenFilenames[3]), self.parameterDict, True)

        if launch:
            self.launchTrue = True

        # Exit with success
        self.accept()

    def closeEvent(self, event):
        self.reject()

    def getLaunch(self):
        return self.launchTrue

    def setWorkspace(self, workspace):
        self.workspace = workspace
        return

def featureExtractionThread(workspace):
    p = subprocess.Popen(["python", "2-featureExtractionInterface.py", fullWorkspacePath],
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    while True:
        output = p.stdout.readline()
        if p.poll() is not None:
            break
        if output:
            print(str(output))
            if "Process finished with exit code" in str(output):
                break
    return

def launchOpenvibe(command, acqScen):
    p = subprocess.Popen([command, "--open", acqScen],
                         stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    while True:
        output = p.stdout.readline()
        if p.poll() is not None:
            break
        if output:
            print(str(output))
            if "Application terminated" in str(output):
                break
    return


def launch(folder, fullWorkspacePath):
    app = QApplication(sys.argv)

    # Check that workspace file exists, is a json file, and contains HappyFeatVersion field...
    if not os.path.exists(fullWorkspacePath):
        myMsgBox("NEED FULL WORKSPACE FILE !!")
        # TODO
        return -1
    with open(fullWorkspacePath, "r") as wp:
        workDict = json.load(wp)
        if not "HappyFeatVersion" in workDict:
            myMsgBox("INVALID WORKSPACE FILE !!")
            # TODO
            return -1

    dlg = Dialog(fullWorkspacePath)
    result = dlg.exec()
    # Dlg exit with error
    if not result:
        return -1

    # Dlg exit success
    if not dlg.getLaunch():
        # No further operation
        text = "Thanks for using the generation script!"
        myMsgBox(text)
        return 0
    else:
        # # Launch Openvibe with acq scenario
        # acqScen = os.path.join(os.getcwd(), "generated", settings.templateScenFilenames[0])
        # command = parameterDict["ovDesignerPath"]
        # threadOV = Thread(target=launchOpenvibe, args=(command, acqScen))
        # threadOV.start()

        # Launch offline extraction interface
        p = Process(target=featExtractApp.launch, args=(folder, fullWorkspacePath))
        p.start()
        p.join()

    return app.exec_()


if __name__ == '__main__':
    retVal = -1
    # Check that a workspace file has been provided
    if len(sys.argv) == 1:
        myMsgBox("NEED WORKSPACE FILE !!")
        # TODO
        sys.exit(-1)

    elif len(sys.argv) == 2:
        folder = os.path.dirname(os.path.abspath(sys.argv[0]))
        retVal = launch(folder, sys.argv[1])

    sys.exit(retVal)



import numpy as np
from scipy import stats

def Compute_Rsquare_Map_optimization(Power_of_trials_1,Power_of_trials_2):
    b = Power_of_trials_1.shape
    a = Power_of_trials_2.shape
    #print(a)
    #print(b)
    Rsquare_tab = np.zeros([a[1]])

    for l in range(b[1]):
        concat_tab_MI = []
        concat_tab_Rest = []
        for i in range(a[0]):
            concat_tab_MI.append(Power_of_trials_1[i,l])
            concat_tab_Rest.append(Power_of_trials_2[i,l])
            #correlation_matrix = np.corrcoef(concat_tab_MI, concat_tab_Rest)
        Sum_q = sum(concat_tab_MI)
        Sum_r = sum(concat_tab_Rest)
        n1 = len(concat_tab_MI)
        n2 = len(concat_tab_Rest)
        sumsqu1 = sum(np.multiply(concat_tab_MI,concat_tab_MI))
        sumsqu2 = sum(np.multiply(concat_tab_Rest,concat_tab_Rest))

        G=((Sum_q+Sum_r)**2)/(n1+n2)

        #correlation_xy = correlation_matrix[0,1]
        #Rsquare_tab[k,l] = correlation_xy**2
        Rsquare_tab[l] = (Sum_q**2/n1+Sum_r**2/n2-G)/(sumsqu1+sumsqu2-G)

    return Rsquare_tab


def Compute_Rsquare_Map(Power_of_trials_1,Power_of_trials_2):
    b = Power_of_trials_1.shape
    a = Power_of_trials_2.shape
    #print(a)
    #print(b)
    Rsquare_tab = np.zeros([a[1],a[2]])
    for k in range(b[1]):
        for l in range(b[2]):
            concat_tab_MI = []
            concat_tab_Rest = []
            for i in range(min(a[0], b[0])):
                concat_tab_MI.append(Power_of_trials_1[i,k,l])
                concat_tab_Rest.append(Power_of_trials_2[i,k,l])
            #correlation_matrix = np.corrcoef(concat_tab_MI, concat_tab_Rest)
            Sum_q = sum(concat_tab_MI)
            Sum_r = sum(concat_tab_Rest)
            n1 = len(concat_tab_MI)
            n2 = len(concat_tab_Rest)
            sumsqu1 = sum(np.multiply(concat_tab_MI,concat_tab_MI))
            sumsqu2 = sum(np.multiply(concat_tab_Rest,concat_tab_Rest))

            G=((Sum_q+Sum_r)**2)/(n1+n2)

            #correlation_xy = correlation_matrix[0,1]
            #Rsquare_tab[k,l] = correlation_xy**2
            Rsquare_tab[k,l] = (Sum_q**2/n1+Sum_r**2/n2-G)/(sumsqu1+sumsqu2-G)

    return Rsquare_tab

def Compute_Wilcoxon_Map_optimization(Power_of_trials_1,Power_of_trials_2):
    b = Power_of_trials_1.shape
    a = Power_of_trials_2.shape

    Wsquare_tab = np.zeros([a[1]])
    Wpsqure_tab = np.zeros([a[1]])
    for k in range(b[1]):
        concat_tab_MI = []
        concat_tab_Rest = []
        for i in range(a[0]):
            concat_tab_MI.append(Power_of_trials_1[i,k])
            concat_tab_Rest.append(Power_of_trials_2[i,k])

        s,p = stats.ranksums(concat_tab_MI,concat_tab_Rest)
        Wsquare_tab[k] = s
        Wpsqure_tab[k] = p
    return Wsquare_tab,Wpsqure_tab

def Compute_Wilcoxon_Map(Power_of_trials_1,Power_of_trials_2):
    b = Power_of_trials_1.shape
    a = Power_of_trials_2.shape

    Wsquare_tab = np.zeros([a[1],a[2]])
    Wpsqure_tab = np.zeros([a[1],a[2]])
    for k in range(b[1]):
        for l in range(b[2]):
            concat_tab_MI = []
            concat_tab_Rest = []
            for i in range(a[0]):
                concat_tab_MI.append(Power_of_trials_1[i,k,l])
                concat_tab_Rest.append(Power_of_trials_2[i,k,l])

            s,p = stats.ranksums(concat_tab_MI,concat_tab_Rest)
            Wsquare_tab[k,l] = s
            Wpsqure_tab[k,l] = p
    return Wsquare_tab,Wpsqure_tab

def Compute_Signed_Rsquare(Power_of_trials_1,Power_of_trials_2):
    b = Power_of_trials_1.shape
    a = Power_of_trials_2.shape
    #print(a)
    #print(b)
    Rsquare_tab = np.zeros([a[1],a[2]])
    Wsquare_tab = np.zeros([a[1],a[2]])
    for k in range(b[1]):
        for l in range(b[2]):
            concat_tab_MI = []
            concat_tab_Rest = []
            for i in range(a[0]):
                concat_tab_MI.append(Power_of_trials_1[i,k,l])
                concat_tab_Rest.append(Power_of_trials_2[i,k,l])
            #correlation_matrix = np.corrcoef(concat_tab_MI, concat_tab_Rest)
            Sum_q = sum(concat_tab_MI)
            Sum_r = sum(concat_tab_Rest)
            n1 = len(concat_tab_MI)
            n2 = len(concat_tab_Rest)
            sumsqu1 = sum(np.multiply(concat_tab_MI,concat_tab_MI))
            sumsqu2 = sum(np.multiply(concat_tab_Rest,concat_tab_Rest))

            G=((Sum_q+Sum_r)**2)/(n1+n2)

            #correlation_xy = correlation_matrix[0,1]
            #Rsquare_tab[k,l] = correlation_xy**2
            s,p = stats.ranksums(concat_tab_MI,concat_tab_Rest)
            
            Wsquare_tab[k,l] = s
            Rsquare_tab[k,l] = ((Sum_q**2/n1+Sum_r**2/n2-G)/(sumsqu1+sumsqu2-G))
    Wsquare_tab = Wsquare_tab/abs(Wsquare_tab)
    return Wsquare_tab*Rsquare_tab


def Compute_Signed_Rsquare_optimization(Power_of_trials_1,Power_of_trials_2):
    b = Power_of_trials_1.shape
    a = Power_of_trials_2.shape
    #print(a)
    #print(b)
    Rsquare_tab = np.zeros([a[1]])
    Wsquare_tab = np.zeros([a[1]])
    for l in range(b[1]):
        concat_tab_MI = []
        concat_tab_Rest = []
        for i in range(a[0]):
            concat_tab_MI.append(Power_of_trials_1[i,l])
            concat_tab_Rest.append(Power_of_trials_2[i,l])
            #correlation_matrix = np.corrcoef(concat_tab_MI, concat_tab_Rest)
        Sum_q = sum(concat_tab_MI)
        Sum_r = sum(concat_tab_Rest)
        n1 = len(concat_tab_MI)
        n2 = len(concat_tab_Rest)
        sumsqu1 = sum(np.multiply(concat_tab_MI,concat_tab_MI))
        sumsqu2 = sum(np.multiply(concat_tab_Rest,concat_tab_Rest))

        G=((Sum_q+Sum_r)**2)/(n1+n2)
        s,p = stats.ranksums(concat_tab_MI,concat_tab_Rest)
        #correlation_xy = correlation_matrix[0,1]
        #Rsquare_tab[k,l] = correlation_xy**2
        Rsquare_tab[l] = np.sign(s)*((Sum_q**2/n1+Sum_r**2/n2-G)/(sumsqu1+sumsqu2-G))
        Wsquare_tab[l] = s
    Wsquare_tab = Wsquare_tab/abs(Wsquare_tab)
    return Wsquare_tab*Rsquare_tab



import mne
import pandas as pd
import matplotlib
from pathlib import Path

df = pd.read_csv('channel_example.txt')
ch_names = df.name.to_list()

pos = df[['x', 'y', 'z']].values
dig_ch_pos = dict(zip(ch_names, pos))
montage = mne.channels.make_dig_montage(ch_pos=dig_ch_pos)
fig = montage.plot()


easycap_montage = mne.channels.make_standard_montage("easycap-M1")
easycap_montage.plot()  # 2D
fig = easycap_montage.plot(kind="3d", show=False)  # 3D
fig = fig.gca().view_init(azim=70, elev=15)  # set view angle for tutorial

a = 0